#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Dict, Optional, Union, Any, Mapping
import enum
from fastapi import Body, Request, HTTPException, Query as FastApiQuery
import pydantic
from pydantic import (  # pylint: disable=unused-import
    BaseModel,
    StrictInt,
    StrictFloat,
    StrictBool,
    Field,
    Extra,
    validator,
    root_validator,
)
from pydantic.main import create_model
import datetime
import numpy as np
import re
import fnmatch
import json

from nomad import datamodel, metainfo  # pylint: disable=unused-import
from nomad.utils import strip
from nomad.metainfo import Datetime, MEnum
from nomad.metainfo.elasticsearch_extension import DocumentType, material_entry_type, material_type

from .utils import parameter_dependency_from_model, update_url_query_arguments


User = datamodel.User.m_def.a_pydantic.model
# It is important that datetime.datetime comes last. Otherwise, number valued strings
# are interpreted as epoch dates by pydantic
Value = Union[StrictInt, StrictFloat, StrictBool, str, datetime.datetime]
ComparableValue = Union[StrictInt, StrictFloat, str, datetime.datetime]


owner_documentation = strip('''
The `owner` allows to limit the scope of the searched based on entry ownership.
This is useful, if you only want to search among all publically downloadable
entries, or only among your own entries, etc.

These are the possible owner values and their meaning:

* `public`: Consider all entries that can be publically downloaded, i.e. only published entries without embargo.
* `user`: Only consider entries that belong to you.
* `shared`: Only consider entries that belong to you or are shared with you.
* `visible`: Consider all entries that are visible to you. This includes
    entries with embargo or unpublished entries that belong to you or are
    shared with you.
* `staging`: Only search through unpublished entries.
* `all`: Consider all entries.
''')


class Owner(str, enum.Enum):
    owner_documentation

    # There seems to be a slight bug in fast API. When it creates the example in OpenAPI
    # it will ignore any given default or example and simply take the first enum value.
    # Therefore, we put public first, which is the most default and save in most contexts.
    public = 'public'
    all_ = 'all'
    visible = 'visible'
    shared = 'shared'
    user = 'user'
    staging = 'staging'
    admin = 'admin'


class Direction(str, enum.Enum):
    '''
    Order direction, either ascending (`asc`) or descending (`desc`)
    '''
    asc = 'asc'
    desc = 'desc'


class HTTPExceptionModel(BaseModel):
    detail: str


class NoneEmptyBaseModel(BaseModel):
    @root_validator
    def check_exists(cls, values):  # pylint: disable=no-self-argument
        assert any(value is not None for value in values.values())
        return values


class All(NoneEmptyBaseModel, extra=Extra.forbid):
    op: List[Value] = Field(None, alias='all')


class None_(NoneEmptyBaseModel, extra=Extra.forbid):
    op: List[Value] = Field(None, alias='none')


class Any_(NoneEmptyBaseModel, extra=Extra.forbid):
    op: List[Value] = Field(None, alias='any')


class Range(BaseModel, extra=Extra.forbid):
    '''
    Represents a finite range which can have open or closed ends. Supports
    several datatypes that have a well-defined comparison operator.
    '''
    @root_validator
    def check_range_is_valid(cls, values):  # pylint: disable=no-self-argument
        lt = values.get('lt')
        lte = values.get('lte')
        gt = values.get('gt')
        gte = values.get('gte')

        # At least one value needs to be defined
        assert (lt is not None) or (lte is not None) or (gt is not None) or (gte is not None)

        # The start/end can only be either open or closed, not both
        if lt is not None:
            assert lte is None
        if lte is not None:
            assert lt is None
        if gt is not None:
            assert gte is None
        if gte is not None:
            assert gt is None

        return values

    lt: Optional[ComparableValue] = Field(None)
    lte: Optional[ComparableValue] = Field(None)
    gt: Optional[ComparableValue] = Field(None)
    gte: Optional[ComparableValue] = Field(None)


ops = {
    'lte': Range,
    'lt': Range,
    'gte': Range,
    'gt': Range,
    'all': All,
    'none': None_,
    'any': Any_
}

CriteriaValue = Union[Value, List[Value], Range, Any_, All, None_, Dict[str, Any]]


class LogicalOperator(NoneEmptyBaseModel):
    @validator('op', check_fields=False)
    def validate_query(cls, query):  # pylint: disable=no-self-argument
        if isinstance(query, list):
            return [_validate_query(item) for item in query]

        return _validate_query(query)


class And(LogicalOperator):
    op: List['Query'] = Field(None, alias='and')


class Or(LogicalOperator):
    op: List['Query'] = Field(None, alias='or')


class Not(LogicalOperator):
    op: 'Query' = Field(None, alias='not')


class Nested(BaseModel):
    prefix: str
    query: 'Query'

    @validator('query')
    def validate_query(cls, query):  # pylint: disable=no-self-argument
        return _validate_query(query)


class Criteria(BaseModel, extra=Extra.forbid):
    name: str
    value: CriteriaValue

    @validator('value')
    def validate_query(cls, value, values):  # pylint: disable=no-self-argument
        name, value = _validate_criteria_value(values['name'], value)
        values['name'] = name
        return value


class Empty(BaseModel, extra=Extra.forbid):
    pass


Query = Union[And, Or, Not, Nested, Criteria, Empty, Mapping[str, CriteriaValue]]


And.update_forward_refs()
Or.update_forward_refs()
Not.update_forward_refs()
Nested.update_forward_refs()


query_documentation = strip('''
A query can be a very simple list of parameters. Different parameters or values of the same parameter are combined
with a logical **and**.
The following query would search for all entries that are VASP calculations,
contain *Na* **and** *Cl*, **and** are authored by *Stefano Curtarolo*
**and** *Chris Wolverton*.
```json
{
    "results.material.elements": ["Na", "Cl"],
    "results.method.simulation.program_name": "VASP",
    "authors": ["Stefano Curtarolo", "Chris Wolverton"]
}
```

A short cut to change the logical combination of values in a list, is to
add a suffix to the quantity `:any`:
```json
{
    "results.material.elements": ["Na", "Cl"],
    "results.method.simulation.program_name": "VASP",
    "authors:any": ["Stefano Curtarolo", "Chris Wolverton"]
}
```

Otherwise, you can also write complex logical combinations of parameters like this:
```json
{
    "and": [
        {
            "or": [
                {
                    "results.material.elements": ["Cl", "Na"]
                },
                {
                    "results.material.elements": ["H", "O"]
                }
            ]
        },
        {
            "not": {
                "results.material.symmetry.crystal_system": "cubic"
            }
        }
    ]
}
```
Other short-cut prefixes are `none:` and `any:` (the default).

By default all quantity values have to **equal** the given values to match. For
some values you can also use comparison operators like this:
```json
{
    "upload_create_time": {
        "gt": "2020-01-01",
        "lt": "2020-08-01"
    },
    "results.properties.geometry_optimization.final_energy_difference": {
        "lte": 1.23e-18
    }
}
```

or shorter with suffixes:
```json
{
    "upload_create_time:gt": "2020-01-01",
    "upload_create_time:lt": "2020-08-01",
    "results.properties.geometry_optimization.final_energy_difference:lte": 1.23e-18
}
```

The searchable quantities are a subset of the NOMAD Archive quantities defined
in the NOMAD Metainfo. The searchable quantities also depend on the API endpoint.

There is also an additional query parameter that you can use to formulate queries based
on the optimade filter language:
```json
{
    "optimade_filter": "nelements >= 2 AND elements HAS ALL 'Ti', 'O'"
}
```
''')


class WithQuery(BaseModel):
    owner: Optional[Owner] = Body('public')
    query: Optional[Query] = Body(
        None,
        embed=True,
        description=query_documentation,
        example={
            'upload_create_time:gt': '2020-01-01',
            'results.material.elements': ['Ti', 'O'],
            'results.method.simulation.program_name': 'VASP',
            'results.properties.geometry_optimization.final_energy_difference:lte': 1.23e-18,
            'results.properties.available_properties': 'section_dos',
            'results.material.type_structural:any': ['bulk', '2d'],
            'optimade_filter': 'nelements >= 2 AND elements HAS ALL "Ti", "O"'
        })

    @validator('query')
    def validate_query(cls, query):  # pylint: disable=no-self-argument
        return _validate_query(query)


def _validate_criteria_value(name: str, value: CriteriaValue):
    if ':' in name:
        quantity, qualifier = name.split(':')
    else:
        quantity, qualifier = name, None

    if qualifier is not None:
        assert qualifier in ops, 'unknown quantity qualifier %s' % qualifier
        return quantity, ops[qualifier](**{qualifier: value})  # type: ignore
    elif isinstance(value, list):
        return quantity, All(all=value)
    else:
        return quantity, value


def _validate_query(query: Query):
    if isinstance(query, dict):
        for key, value in list(query.items()):
            quantity, value = _validate_criteria_value(key, value)
            if quantity != key:
                assert quantity not in query, 'a quantity can only appear once in a query'
                del(query[key])
            query[quantity] = value

    return query


class QueryParameters:
    def __init__(self, doc_type: DocumentType):
        self.doc_type = doc_type

    def __call__(
        self,
        request: Request,
        owner: Optional[Owner] = FastApiQuery(
            'public', description=strip(Owner.__doc__)),
        json_query: Optional[str] = FastApiQuery(None, description=strip('''
                To pass a query string in the format of JSON e.g. '{{"results.material.elements": ["H", "O"]}}'.
            ''')),
        q: Optional[List[str]] = FastApiQuery(
            [], description=strip('''
                Since we cannot properly offer forms for all parameters in the OpenAPI dashboard,
                you can use the parameter `q` and encode a query parameter like this
                `atoms__H` or `n_atoms__gt__3`. Multiple usage of `q` will combine parameters with
                logical *and*.
            '''))) -> WithQuery:

        # copy parameters from request
        query_params = {
            key: request.query_params.getlist(key)
            for key in request.query_params}

        # add the encoded parameters
        for parameter in q:
            fragments = parameter.split('__')
            if len(fragments) == 1 or len(fragments) > 3:
                raise HTTPException(422, detail=[{
                    'loc': ['query', 'q'],
                    'msg': 'wrong format, use <quantity>[__<op>]__<value>'}])
            name_op, value = '__'.join(fragments[:-1]), fragments[-1]
            quantity_name = name_op.split('__')[0]

            doc_type = self.doc_type
            if quantity_name.startswith('entries.'):
                if self.doc_type == material_type:
                    doc_type = material_entry_type
                else:
                    raise HTTPException(422, detail=[{
                        'loc': ['query', parameter],
                        'msg': f'entries can only be nested into material queries'}])

            if quantity_name not in doc_type.quantities:
                raise HTTPException(422, detail=[{
                    'loc': ['query', parameter],
                    'msg': f'{quantity_name} is not a {doc_type} quantity'}])

            query_params.setdefault(name_op, []).append(value)

        # transform query parameters to query
        query: Dict[str, Any] = {}
        for key in query_params:
            op = None
            if '__' in key:
                quantity_name, op = key.split('__')
            else:
                quantity_name = key

            if quantity_name.startswith('entries.'):
                quantity = material_entry_type.quantities.get(quantity_name[8:])
            else:
                quantity = self.doc_type.quantities.get(quantity_name)

            if quantity is None:
                continue

            type_ = quantity.definition.type
            if type_ is Datetime:
                type_ = datetime.datetime.fromisoformat
            elif isinstance(type_, MEnum):
                type_ = str
            elif isinstance(type_, np.dtype):
                type_ = float
            elif type_ not in [int, float, bool]:
                type_ = str
            values = query_params[key]
            values = [type_(value) for value in values]

            if op is None:
                op = 'all' if quantity.many_all else 'any'

            if op == 'all':
                query[quantity_name] = All(all=values)
            elif op == 'any':
                query[quantity_name] = Any_(any=values)
            elif op in ops:
                if len(values) > 1:
                    raise HTTPException(
                        status_code=422,
                        detail=[{
                            'loc': ['query', key],
                            'msg': 'operator %s does not support multiple values' % op}])
                query[quantity_name] = ops[op](**{op: values[0]})
            else:
                raise HTTPException(
                    422, detail=[{'loc': ['query', key], 'msg': 'operator %s is unknown' % op}])

        # process the json_query
        if json_query is not None:
            try:
                query.update(**json.loads(json_query))
            except Exception:
                raise HTTPException(
                    422, detail=[{'loc': ['json_query'], 'msg': 'cannot parse json_query'}])

        return WithQuery(query=query, owner=owner)


class MetadataRequired(BaseModel):
    ''' Defines which metadata quantities are included or excluded in the response. '''

    include: Optional[List[str]] = Field(
        None, description=strip('''
            Quantities to include for each result. Only those quantities will be
            returned. At least one id quantity (e.g. `entry_id`) will always be included.
        '''))
    exclude: Optional[List[str]] = Field(
        None, description=strip('''
            Quantities to exclude for each result. Only all other quantities will
            be returned. The entity's id quantity (e.g. `entry_id`) cannot be excluded.
        '''))


metadata_required_parameters = parameter_dependency_from_model(
    'metadata_required_parameters', MetadataRequired)


class Pagination(BaseModel):
    ''' Defines the order, size, and page of results. '''

    page_size: Optional[int] = Field(
        10, description=strip('''
            The page size, e.g. the maximum number of items contained in one response.
            A `page_size` of 0 will return no results.
        '''))
    order_by: Optional[str] = Field(
        None,
        description=strip('''
            The results are ordered by the values of this field. If omitted, default
            ordering is applied.
        '''))
    order: Optional[Direction] = Field(
        Direction.asc, description=strip('''
            The ordering direction of the results based on `order_by`. Its either
            ascending `asc` or decending `desc`. Default is `asc`.
        '''))
    page_after_value: Optional[str] = Field(
        None, description=strip('''
            This attribute defines the position after which the page begins, and is used
            to navigate through the total list of results.

            When requesting the first page, no value should be provided for
            `page_after_value`. Each response will contain a value `next_page_after_value`,
            which can be used to obtain the next page (by setting `page_after_value` in
            your next request to this value).

            The field is encoded as a string, and the format of `page_after_value` and
            `next_page_after_value` depends on which API method is used.

            Some API functions additionally allows a simplified navigation, by specifying
            the page number in the key `page`. It is however always possible to use
            `page_after_value` and `next_page_after_value` to iterate through the results.
            '''))
    page: Optional[int] = Field(
        None, description=strip('''
            The number of the page (1-based). When provided in a request, this attribute
            can be used instead of `page_after_value` to jump to a particular results page.

            **NOTE #1**: the option to request pages by submitting the `page` number is
            limited. There are api calls where this attribute cannot be used for indexing,
            or where it can only be used partially. **If you want to just iterate through
            all the results, aways use the `page_after_value` and `next_page_after_value`!**

            **NOTE #2**: Only one, `page`, `page_offset` or `page_after_value`, can be used.
        '''))
    page_offset: Optional[int] = Field(
        None, description=strip('''
            The number of skipped entries. When provided in a request, this attribute
            can be used instead of `page_after_value` to jump to a particular results page.

            **NOTE #1**: the option to request pages by submitting the `page_offset` number is
            limited. There are api calls where this attribute cannot be used for indexing,
            or where it can only be used partially. **If you want to just iterate through
            all the results, aways use the `page_after_value` and `next_page_after_value`!**

            **NOTE #2**: Only one, `page`, `page_offset` or `page_after_value`, can be used.
        '''))

    @validator('page_size')
    def validate_page_size(cls, page_size):  # pylint: disable=no-self-argument
        assert page_size >= 0, 'page_size must be >= 0'
        return page_size

    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        '''
        Override this in your Pagination class to ensure that a valid attribute is selected.
        This method has to be implemented!
        '''
        raise NotImplementedError('Validation of `order_by` not implemented!')

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        '''
        Override this in your Pagination class to implement validation of the
        `page_after_value` value.
        This method has to be implemented!
        '''
        raise NotImplementedError('Validation of `page_after_value` not implemented!')

    @validator('page')
    def validate_page(cls, page, values):  # pylint: disable=no-self-argument
        if page is not None:
            assert page >= 1, 'page must be >= 1'
        return page

    @validator('page_offset')
    def validate_page_offset(cls, page_offset, values):  # pylint: disable=no-self-argument
        if page_offset is not None:
            assert page_offset >= 0, 'page_offset must be >= 1'
        return page_offset

    @root_validator(skip_on_failure=True)
    def validate_values(cls, values):  # pylint: disable=no-self-argument
        # Because of a bug in pydantic (#2670), root validators can't be overridden, so
        # we invoke a class method, which *can* be overridden.
        return cls._root_validation(values)

    @classmethod
    def _root_validation(cls, values):
        page_offset = values.get('page_offset')
        page = values.get('page')
        page_after_value = values.get('page_after_value')
        page_size = values.get('page_size')

        n_offset_criteria = (1 if page_offset else 0) + (1 if page else 0) + (1 if page_after_value else 0)
        assert n_offset_criteria <= 1, 'Can only specify one `page_offset`, `page`, or `page_after_value'

        if page_size == 0:
            assert page_offset is None, 'Cannot specify `page_offset` when `page_size` is set to 0'
            assert page is None, 'Cannot specify `page` when `page_size` is set to 0'
            assert page_after_value is None, 'Cannot specify `page_after_value` when `page_size` is set to 0'

        return values

    def get_simple_index(self):
        '''
        If simple, index-based pagination is used, this method can be used to get the
        corresponding index (0-based). It will look on either `page` or `page_after_value`.
        If neither index is provided, we return 0 (i.e. the first index).
        '''
        if self.page_offset is not None:
            return self.page_offset
        if self.page is not None:
            return (self.page - 1) * self.page_size
        if self.page_after_value is not None:
            rv = int(self.page_after_value) + 1
            assert rv >= 0
            return rv
        return 0


class PaginationResponse(Pagination):
    total: int = Field(
        ..., description=strip('''
        The total number of results that fit the given query. This is independent of
        any pagination and aggregations.
        '''))
    next_page_after_value: Optional[str] = Field(
        None, description=strip('''
        The *next* value to be used as `page_after_value` in a follow up requests, to get
        the next page of results. If no more results are available, `next_page_after_value`
        will not be set.
        '''))
    page_url: Optional[str] = Field(
        None, description=strip('''
        The url of the current page. Only applicable for GET requests.
        '''))
    next_page_url: Optional[str] = Field(
        None, description=strip('''
        The url to get the next page. Only applicable for GET requests.
        '''))
    prev_page_url: Optional[str] = Field(
        None, description=strip('''
        The url to get the previous page. **NOTE:** Only applicable for some API methods,
        (namely, where indexing by `page` is possible), and only for GET requests.
        '''))
    first_page_url: Optional[str] = Field(
        None, description=strip('''
        The url to get the first page. Only applicable for GET requests.
        '''))

    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        # No validation - behaviour of this field depends on api method
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # No validation - behaviour of this field depends on api method
        return page_after_value

    @classmethod
    def _root_validation(cls, values):  # pylint: disable=no-self-argument
        # No validation
        return values

    def populate_urls(self, request: Request):
        '''
        Populates the urls (`page_url`, `next_page_url`, `first_page_url` from the
        request and `next_page_after_value`. Only applicable for GET requests.
        '''
        assert request.method.upper() == 'GET', 'Trying to populate urls, but method is not GET.'
        original_url = str(request.url)
        self.page_url = original_url
        if self.page_size:
            self.first_page_url = update_url_query_arguments(
                original_url, page=None, page_after_value=None)
        if self.next_page_after_value:
            self.next_page_url = update_url_query_arguments(
                original_url, page=None, page_after_value=self.next_page_after_value)
        if self.page and self.page > 1:
            self.prev_page_url = update_url_query_arguments(
                original_url, page=self.page - 1, page_after_value=None)

    def populate_simple_index_and_urls(self, request: Request):
        '''
        If simple, index-based pagination is used, this method can be used to populate
        the `page`, `page_after_value` and urls (if it is a GET request) automatically.
        Assumes that the field `total` is populated.
        '''
        if not self.page_size:
            self.page = 1
            self.page_after_value = None
            self.next_page_after_value = None
        else:
            ind = self.get_simple_index()
            self.page = ind // self.page_size + 1
            self.page_after_value = None if self.page == 1 else str(ind - 1)
            if self.page_size * self.page >= self.total:
                self.next_page_after_value = None
            else:
                self.next_page_after_value = str(ind + self.page_size - 1)

            if self.page < 1 or (
                    self.total == 0 and self.page != 1) or (
                    self.total > 0 and (self.page - 1) * self.page_size >= self.total):
                raise HTTPException(400, detail='Page out of range requested.')
        if request.method.upper() == 'GET':
            self.populate_urls(request)


class MetadataBasedPagination(Pagination):
    order_by: Optional[str] = Field(
        None,
        description=strip('''
            The results are ordered by the values of this field. If omitted, default
            ordering is applied.
        '''))

    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        # No validation here – validation is done during search
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # No validation here – validation is done during search
        return page_after_value


class MetadataPagination(MetadataBasedPagination):
    page: Optional[int] = Field(
        None, description=strip('''
            For simple, index-based pagination, this should contain the number of the
            requested page (1-based). When provided in a request, this attribute can be
            used instead of `page_after_value` to jump to a particular results page.

            However, you can only retreive up to the 10.000th entry with a page number.
            Only one, `page`, `page_offset` or `page_after_value`, can be used.
        '''))

    page_offset: Optional[int] = Field(
        None, description=strip('''
            Return the page that follows the given number of entries. Overwrites
            `page` and `page_after_value`.

            However, you can only retreive up to the 10.000th entry.
            Only one, `page`, `page_offset` or `page_after_value`, can be used.
        ''')
    )

    @validator('page')
    def validate_page(cls, page, values):  # pylint: disable=no-self-argument
        if page is not None:
            assert page > 0, 'Page has to be larger than 1.'
            assert page * values.get('page_size', 10) < 10000, 'Pagination by `page` is limited to 10.000 entries.'

        return page

    @validator('page_offset')
    def validate_page_offset(cls, page_offset, values):  # pylint: disable=no-self-argument
        if page_offset is not None:
            assert page_offset >= 0, 'Page offset has to be larger than 0.'
            assert page_offset + values.get('page_size', 10) < 10000, 'Page offset plus page size has to be smaller thant 10.0000.'

        return page_offset


metadata_pagination_parameters = parameter_dependency_from_model(
    'metadata_pagination_parameters', MetadataPagination)


class AggregationPagination(MetadataBasedPagination):
    order_by: Optional[str] = Field(
        None,  # type: ignore
        description=strip('''
            Either the string "count", "value", or the name of a quantity. If omitted the buckets
            will be ordered by the item "count".

            If you provide a quantity, all items
            in a bucket must have the same value for this quantity. For example, aggregating
            entries on `upload_id` and ordering with the buckets by `upload_create_time` is fine,
            because all entries in an upload have the same `upload_create_time`. The API cannot
            check this rule and the results will be unpredictable.

            If you want to order by the bucket values, you can either use "value" or use
            the aggregation quantity to `order_by`. The result will be the same, because
            the bucket values are the quantity values.
        '''))

    @validator('page')
    def validate_page(cls, page, values):  # pylint: disable=no-self-argument
        assert page is None, 'Pagination by `page` is not possible for aggregations, use `page_after_value`'
        return page

    @validator('page_size')
    def validate_page_size(cls, page_size, values):  # pylint: disable=no-self-argument
        assert page_size > 0, '0 or smaller page sizes are not allowed for aggregations.'
        return page_size


class AggregatedEntities(BaseModel):
    size: Optional[pydantic.conint(gt=0)] = Field(  # type: ignore
        1, description=strip('''
        The number of entries that should be returned for each value in the
        aggregation.
        '''))
    required: Optional[MetadataRequired] = Field(
        None, description=strip('''
        This allows to determined what fields should be returned for each entry.
        '''))


class AggregationBase(BaseModel):
    pass


class QuantityAggregation(AggregationBase):
    quantity: str = Field(
        ..., description=strip('''
        The manatory name of the quantity for the aggregation. Aggregations
        can only be computed for those search metadata that have discrete values;
        an aggregation buckets entries that have the same value for this quantity.'''))
    exclude_from_search: bool = Field(
        False, description=strip('''
        If set to true, top-level search criteria involving the aggregation quantity, will not
        be applied for this aggregation. Therefore, the aggregation will return all
        values for the quantity, even if the possible values where filtered by the query.

        There are two limitations. This is only supported with queries that start with a
        dictionary. It will not work for queries with a boolean operator. It can only
        exclude top-level criteria at the root of the query dictionary. Nested criteria,
        e.g. within complex and/or constructs, cannot be considered. Using this might also
        prohibit pagination with page_after_value on aggregations in the same request.
        ''')
    )


class BucketAggregation(QuantityAggregation):
    metrics: Optional[List[str]] = Field(
        [], description=strip('''
        By default the returned aggregations will provide the number of entries for each
        value. You can add more metrics. For each metric an additional number will be
        provided for each value. Metrics are also based on search metadata. Depending on
        the metric the number will represent either a sum (`calculations` for the number
        of individual calculation in each code run) or an amount of different values
        (i.e. `materials` for the amount of different material hashes).'''))


class TermsAggregation(BucketAggregation):
    pagination: Optional[AggregationPagination] = Field(
        None, description=strip('''
        Only the data few values are returned for each API call. Aggregation
        pagination allows to get all available values by pagination. It also allows to
        order values.

        You can only use pagination (to page through all available values) or size (to
        get the size number of values with the most available data).
        '''))
    size: Optional[pydantic.conint(gt=0)] = Field(  # type: ignore
        None, description=strip('''
        The ammount of aggregation values is limited. This allows you to configure the
        maximum number of aggregated values to return. If you need to exaust all
        possible value, use `pagination`.
        '''))
    include: Optional[Union[List[str], pydantic.constr(regex=r'^[a-zA-Z0-9_\-\s]+$')]] = Field(  # type: ignore
        None, description=strip('''
        An optional filter for aggregation values. You can either specify a
        single string which must be contained in the aggregation value or then
        provide an array of keywords for which the aggregation will be created.

        This is only available for non paginated aggregations.
        '''))
    entries: Optional[AggregatedEntities] = Field(
        None, description=strip('''
        Optionally, a set of entries can be returned for each value. These are basically
        example entries that have the respective bucket value.
        '''))


class HistogramAggregation(BucketAggregation):
    interval: pydantic.confloat(gt=0)  # type: ignore


class DateHistogramAggregation(BucketAggregation):
    interval: str = Field('1M')  # type: ignore


class MinMaxAggregation(QuantityAggregation):
    pass


class StatisticsAggregation(AggregationBase):
    metrics: Optional[List[str]] = Field(
        [], description=strip('''
        A list of search quantities to act as metrics on all data. Depending on
        the metric the number will represent either a sum (`calculations` for the number
        of individual calculation in each code run) or an amount (cardinality) of
        different values (i.e. `materials` for the amount of different material hashes).'''))


class Aggregation(BaseModel):
    terms: Optional[TermsAggregation] = Body(
        None,
        description=strip('''
            A `terms` aggregation allows to get the values of a quantity that occur in
            the search query result data. For each value, a bucket is created with
            information about how many entries have the value (or additional metrics).
            For example to get all entries that use a certain code, you can use:
            ```json
            {
                "aggregations": {
                    "all_codes": {
                        "terms": {
                            "quantity": "results.method.simulation.program_name"
                        }
                    }
                }
            }
            ```

            Terms aggregations can also be used to paginate though all values of a certain
            quantities. Each page will be companied with a `page_after_value` that
            can be used to retrieve the next value. For example to go through all datasets
            available in the search query:
            ```json
            {
                "aggregations": {
                    "all_datasets": {
                        "terms": {
                            "quantity": "datasets",
                            "pagination": {
                                "page_size": 100
                            }
                        }
                    }
                }
            }
            ```
        '''))

    histogram: Optional[HistogramAggregation] = Body(
        None,
        description=strip('''
            A `histogram` aggregation allows to get the ranges quantity values and
            how many (or another metrics) entries exhibit values in those ranges:

            ```json
            {
                "aggregations": {
                    "calculations_per_entry": {
                        "histogram": {
                            "quantity": "properties.n_calculations",
                            "interval": 10
                        }
                    }
                }
            }
            ```

            The used quantity must be a float or int typed quantity. An interval is
            mandatory and determines the bucket size.
        '''))

    date_histogram: Optional[DateHistogramAggregation] = Body(
        None,
        description=strip('''
            A `date_histogram` aggregation is like a normal `histogram` but for date valued
            quantities.:

            ```json
            {
                "aggregations": {
                    "upload_create_times": {
                        "histogram": {
                            "quantity": "upload_create_time",
                            "interval": "1M"
                        }
                    }
                }
            }
            ```

            The used quantity must be a datetime typed quantity. Intervals are strings
            that determine a time period. The default is one month, `1M`.
        '''))

    min_max: Optional[MinMaxAggregation] = Body(
        None,
        description=strip('''
            A `min_max` aggregation allows to get the minumum and maximum quantity values:

            ```json
            {
                "aggregations": {
                    "calculations_per_entry": {
                        "min_max": {
                            "quantity": "results.properties.n_calculations"
                        }
                    }
                }
            }
            ```

            The used quantity must be a float or int typed quantity.
        '''))

    statistics: Optional[StatisticsAggregation] = Body(
        None,
        description=strip('''
            A `statistics` aggregation allows to get metrics (sums or cardinalities) from all data
            that matches the search.

            ```json
            {
                "aggregations": {
                    "statistics": {
                        "global": {
                            "metrics": ["results.properties.n_calculations", "results.material.material_id"]
                        }
                    }
                }
            }
            ```
        '''))


class WithQueryAndPagination(WithQuery):
    pagination: Optional[MetadataPagination] = Body(
        None,
        example={
            'page_size': 5,
            'order_by': 'upload_create_time'
        })


class Metadata(WithQueryAndPagination):
    required: Optional[MetadataRequired] = Body(
        None,
        example={
            'include': ['entry_id', 'mainfile', 'upload_id', 'authors', 'upload_create_time']
        })
    aggregations: Optional[Dict[str, Aggregation]] = Body(
        {},
        example={
            'all_codes': {
                'terms': {
                    'quantity': 'results.method.simulation.program_name',
                    'entries': {
                        'size': 1,
                        'required': {
                            'include': ['mainfile']
                        }
                    }
                },
            },
            'all_datasets': {
                'terms': {
                    'quantity': 'datasets.dataset_name',
                    'pagination': {
                        'page_size': 100
                    }
                }
            },
            'system_size': {
                'min_max': {
                    'quantity': 'results.properties.structures.structure_conventional.n_sites'
                }
            },
            'upload_create_times': {
                'date_histogram': {
                    'quantity': 'upload_create_time',
                    'interval': '1M'
                }
            },
            'calculations_per_entry': {
                'histogram': {
                    'quantity': 'results.properties.n_calculations',
                    'interval': 5
                }
            }
        },
        description=strip('''
            Defines additional aggregations to return. There are different types of
            aggregations: `terms`, `histogram`, `data_histogram`, `min_max`.
        '''))


class MetadataEditListAction(BaseModel):
    '''
    Defines an action to perform on a list quantity. This enables users to add and remove values.
    '''
    set: Optional[Union[str, List[str]]] = Field(description=strip('''
        Value(s) to set. Note, a set-operation overwrites the old list with the provided list.
        If a set-operation is specified, it is therefore not possible to also specify an
        add- or remove-operation.'''))
    add: Optional[Union[str, List[str]]] = Field(description=strip('''
        Value(s) to add to the list'''))
    remove: Optional[Union[str, List[str]]] = Field(description=strip('''
        Value(s) to remove from the list'''))


# Generate model for MetadataEditActions
_metadata_edit_actions_fields = {}
for quantity in datamodel.EditableUserMetadata.m_def.definitions:
    if quantity.is_scalar:
        pydantic_type = quantity.type if quantity.type in (str, int, float, bool) else str
    else:
        pydantic_type = Union[str, List[str], MetadataEditListAction]
    if getattr(quantity, 'a_auth_level', None) == datamodel.AuthLevel.admin:
        description = '**NOTE:** Only editable by admin user'
    else:
        description = None
    _metadata_edit_actions_fields[quantity.name] = (
        Optional[pydantic_type], Field(description=description))

MetadataEditActions = create_model('MetadataEditActions', **_metadata_edit_actions_fields)  # type: ignore


class MetadataEditRequest(WithQuery):
    ''' Defines a request to edit metadata. '''
    metadata: Optional[MetadataEditActions] = Field(  # type: ignore
        description=strip('''
            Metadata to set, on the upload and/or selected entries.'''))
    entries: Optional[Dict[str, MetadataEditActions]] = Field(  # type: ignore
        description=strip('''
            An optional dictionary, specifying metadata to set on individual entries. The field
            `entries_metadata_key` defines which type of key is used in the dictionary to identify
            the entries. Note, only quantities defined on the entry level can be set using this method.'''))
    entries_key: Optional[str] = Field(
        default='entry_id', description=strip('''
            Defines which type of key is used in `entries_metadata`. Default is `entry_id`.'''))
    verify_only: Optional[bool] = Field(
        default=False, description=strip('''
            Do not execute the request, just verifies it and provides detailed feedback on
            encountered errors etc.'''))


class Files(BaseModel):
    ''' Configures the download of files. '''
    compress: Optional[bool] = Field(
        False, description=strip('''
        By default the returned zip file is not compressed. This allows to enable compression.
        Compression will reduce the rate at which data is provided, often below
        the rate of the compression. Therefore, compression is only sensible if the
        network connection is limited.'''))
    glob_pattern: Optional[str] = Field(
        None, description=strip('''
        An optional *glob* (or unix style path) pattern that is used to filter the
        returned files. Only files matching the pattern are returned. The pattern is only
        applied to the end of the full path. Internally
        [fnmatch](https://docs.python.org/3/library/fnmatch.html) is used.'''))
    re_pattern: Optional[str] = Field(
        None, description=strip('''
        An optional regexp that is used to filter the returned files. Only files matching
        the pattern are returned. The pattern is applied in search mode to the full
        path of the files. With `$` and `^` you can control if you want to match the
        whole path.

        A re pattern will replace a given glob pattern.'''))
    include_files: Optional[List[str]] = Field(
        None, description=strip('''
        Optional list of file names. Only files with these names are included in the
        results. This will overwrite any given glob or re pattern.''')
    )

    @validator('glob_pattern')
    def validate_glob_pattern(cls, glob_pattern):  # pylint: disable=no-self-argument
        # compile the glob pattern into re
        if glob_pattern is None:
            return None

        return re.compile(fnmatch.translate(glob_pattern) + r'$')

    @validator('re_pattern')
    def validate_re_pattern(cls, re_pattern):  # pylint: disable=no-self-argument
        # compile an re
        if re_pattern is None:
            return None
        try:
            return re.compile(re_pattern)
        except re.error as e:
            assert False, 'could not parse the re pattern: %s' % e

    @root_validator()
    def vaildate(cls, values):  # pylint: disable=no-self-argument
        # use the compiled glob pattern as re
        if values.get('re_pattern') is None:
            values['re_pattern'] = values.get('glob_pattern')

        if values.get('include_files') is not None:
            files = values['include_files']
            values['re_pattern'] = re.compile(f'({"|".join([re.escape(f) for f in files])})$')

        return values


files_parameters = parameter_dependency_from_model(
    'files_parameters', Files)


class Bucket(BaseModel):
    entries: Optional[List[Dict[str, Any]]] = Field(
        None, description=strip('''The entries that were requested for each value.'''))
    count: int = Field(
        None, description=strip('''The amount of entries with this value.'''))
    nested_count: int = Field(
        None, description=strip('''
            The amount of nested entries with this values. Is the same as count for
            aggregations on non nested quantities.'''))
    metrics: Optional[Dict[str, int]]

    value: Union[StrictBool, float, str]


class BucketAggregationResponse(BaseModel):
    data: List[Bucket] = Field(
        None, description=strip('''The aggregation data as a list.'''))


class TermsAggregationResponse(BucketAggregationResponse, TermsAggregation):
    pagination: Optional[PaginationResponse]  # type: ignore


class HistogramAggregationResponse(BucketAggregationResponse, HistogramAggregation):
    pass


class DateHistogramAggregationResponse(BucketAggregationResponse, DateHistogramAggregation):
    pass


class MinMaxAggregationResponse(MinMaxAggregation):
    data: List[Union[float, None]]


class StatisticsAggregationResponse(StatisticsAggregation):
    data: Optional[Dict[str, int]]


class AggregationResponse(Aggregation):
    terms: Optional[TermsAggregationResponse]
    histogram: Optional[HistogramAggregationResponse]
    date_histogram: Optional[DateHistogramAggregationResponse]
    min_max: Optional[MinMaxAggregationResponse]
    statistics: Optional[StatisticsAggregationResponse]


class CodeResponse(BaseModel):
    curl: str
    requests: str
    nomad_lab: Optional[str]


class MetadataResponse(Metadata):
    pagination: PaginationResponse = None  # type: ignore
    aggregations: Optional[Dict[str, AggregationResponse]]  # type: ignore

    data: List[Dict[str, Any]] = Field(
        None, description=strip('''
        The entries data as a list. Each item is a dictionary with the metadata for each
        entry.'''))

    code: Optional[CodeResponse]
    es_query: Any = Field(
        None, description=strip('''The elasticsearch query that was used to retrieve the results.'''))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re
from datetime import datetime
from ase.io import read as ioread
from ase import Atoms

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.method import (
    Electronic, Method, DFT, Smearing, XCFunctional, Functional, KMesh, BasisSet
)
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Forces, ForcesEntry, ScfIteration, Energy, EnergyEntry, BandEnergies, Dos,
    DosValues
)
from nomad.datamodel.metainfo.workflow import Workflow
from .metainfo.wien2k import x_wien2k_section_equiv_atoms


class In0Parser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity(
                'xc_functional',
                r'(?:TOT|KXC|POT|MULT|COUL|EXCH)\s*([\w ]+)', dtype=str),
            Quantity('fft', r'FFT[\s\S]+?(\d+\s+\d+\s+\d+\s+[\d\.]+)')]


class In1Parser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity('wf_switch', r'(WFFIL|WFPRI|ENFIL|SUPWF)'),
            Quantity('rkmax', r'([\d\.]+\s*\d+\s*\d+).+K\-MAX', dtype=np.float64)]


class StructParser(TextParser):
    def __init__(self):
        super().__init__()
        self._units_map = {
            'b': ureg.bohr, 'a': ureg.angstrom}

    def init_quantities(self):
        re_float = r'[\d\.\-]+'
        re_lat = r'\d+\.\d{6}'
        self._quantities = [
            Quantity(
                'lattice',
                r'([\s\S]+?)\n *AT',
                sub_parser=TextParser(quantities=[
                    Quantity('nonequiv_atoms', r'NONEQUIV\.ATOMS\:\s*(\d+)', dtype=int),
                    Quantity('lattice', r'(\w+)\s*LATTICE'),
                    Quantity('calc_mode', r'(N*REL\S*)'),
                    Quantity(
                        'lattice_constants',
                        # fixed precision, sometimes no spaces
                        rf'({re_lat})\s*({re_lat})\s*({re_lat})\s*({re_lat})\s*({re_lat})\s*({re_lat})',
                        dtype=np.dtype(np.float64)),
                    Quantity(
                        'unit',
                        r'unit=(\w)',
                        str_operation=lambda x: self._units_map.get(x, ureg.bohr))])),
            Quantity(
                'atom',
                r'OM\s+\-*\d+\:\s*(X\=[\s\S]+?)LOCAL',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'positions',
                        rf'X\=({re_float})\s*Y=({re_float})\s*Z=({re_float})',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity('atom_name', r'(\n *[A-Z][a-z]*\d* +)'),
                    Quantity('NPT', r'NPT\s*\=\s*(\d+)', dtype=int),
                    Quantity('R0', r'R0\s*\=\s*(\d+)', dtype=int),
                    Quantity('Z', r'Z\:\s*([\d\.]+)', dtype=np.float64)]))]

    def get_atoms(self):
        '''
        Returns an ASE atoms representation of the structure.
        '''
        if self.mainfile is None:
            return

        try:
            return ioread(self.mainfile, format='struct')
        except Exception:
            # read it from parsed info
            if self.get('lattice') is None:
                return

            lattice_constants = self.get('lattice').get('lattice_constants')
            unit = self.get('lattice').get('unit', ureg.bohr)
            lattice_constants[:3] = (lattice_constants[:3] * unit).to('angstrom').magnitude
            scaled_positions = []
            numbers = []
            for atom in self.get('atom', []):
                positions = atom.get('positions', [])
                scaled_positions.extend(positions)
                numbers.extend([int(atom.get('Z', 0))] * len(positions))

            return Atoms(
                cell=lattice_constants, scaled_positions=scaled_positions, numbers=numbers, pbc=True)


class In2Parser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity('switch', r'(TOT|FOR|QTL|EFG|FERMI)'),
            Quantity('emin', r'([\d\.\- ]+)\s*EMIN'),
            Quantity('smearing', r'(GAUSS|ROOT|TEMP|TETRA|ALL)\s*([\d\.]+)'),
            Quantity('gmax', r'([\d\.\-]+)\s*GMAX')]


class DosParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_f = r'\-*\d+\.\d+'
        self._quantities = [
            Quantity('labels', r'# ENERGY +(.+)', flatten=False),
            Quantity('data', rf'({re_f} +{re_f}.*)', repeats=True, dtype=np.dtype(np.float64))]


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.Ee\-\+]+'

        iteration_quantities = [
            Quantity(
                'NATO',
                r'(NATO\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'nr_of_independent_atoms',
                        r'(\d+)\s*INDEPENDENT', dtype=int),
                    Quantity(
                        'total_atoms',
                        r'(\d+)\s*TOTAL ATOMS IN UNITCELL', dtype=int),
                    Quantity('system_name', r'SUBSTANCE:\s*(.+)', flatten=False)])),
            Quantity(
                'POT',
                r'(POT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'potential_option',
                        r'POTENTIAL OPTION\s*(.+)', dtype=str, flatten=False)])),
            Quantity(
                'LAT',
                r'(LAT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity('lattice_const', r'LATTICE CONSTANTS=\s*([\d\. ]+)')])),
            Quantity(
                'VOL',
                r'(VOL\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'unit_cell_volume_bohr3',
                        rf'UNIT CELL VOLUME\s*\=\s*({re_float})',
                        dtype=np.float64),
                    Quantity('spinpolarization', r'((?:NON-)*SPINPOLARIZED) CALCULATION')])),
            Quantity(
                'RKM',
                r'(RKM\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity('matrix_size', r'MATRIX SIZE\s*(\d+)', dtype=int),
                    Quantity('LOs', r'LOs:\s*(\d+)', dtype=int),
                    Quantity('rkm', r'RKM\=\s*([\d\.]+)', dtype=np.float64)])),
            Quantity(
                'KPT',
                r'(KPT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[Quantity(
                    'nr_kpts', r'NUMBER OF K-POINTS:\s*(\d+)', dtype=int)])),
            Quantity(
                'GAP',
                r'(GAP\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[Quantity(
                    'ene_gap', rf'({re_float})\s*Ry', dtype=np.float64, unit=ureg.rydberg)])),
            Quantity(
                'NOE',
                r'(NOE\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'noe',
                        rf'NUMBER OF ELECTRONS\s*\=\s*({re_float})',
                        dtype=np.float64)])),
            Quantity(
                'FER',
                r'(FER\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_reference_fermi',
                        rf'F E R M I \- ENERGY.+?\=\s*([\d\.\-\+Ee ]+)',
                        str_operation=lambda x: [float(v) for v in x.strip().split()] * ureg.rydberg,
                        convert=False)])),
            Quantity(
                'GMA',
                r'(GMA\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'cutoff',
                        rf'POTENTIAL AND CHARGE CUT\-OFF\s*({re_float})',
                        dtype=np.float64)])),
            Quantity(
                'POSi',
                r'(POS\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity('atom_mult', r'MULT.*?\s*\=\s*(\d+)', dtype=int),
                    Quantity(
                        'position',
                        rf'POSITION\s*\=\s*({re_float}\s*{re_float}\s*{re_float})',
                        dtype=np.float64)])),
            Quantity(
                'CHAi',
                r'(CHA\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_val_charge_cell',
                        rf'TOTAL .+?CHARGE INSIDE.+?\=\s*({re_float})',
                        dtype=np.float64)])),
            Quantity(
                'SUM',
                r'(SUM\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_sum_eigenvalues',
                        rf'SUM OF EIGENVALUES\s*\=\s*({re_float})',
                        dtype=np.float64, unit=ureg.rydberg)])),
            Quantity(
                'RTOi',
                rf'RTO\d+\:\s*\d+\s*({re_float})\s*({re_float})\s*({re_float})\s*({re_float})\s*',
                dtype=np.dtype(np.float64), repeats=True),
            Quantity(
                'NTO',
                r'(NTO\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_int_charge_nm',
                        rf'CHARGE\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'NTOi',
                r'(NTO\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_charge_in_sphere_nm',
                        rf'CHARGE.+\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'DTOi',
                r'(DTO\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_diff_charge',
                        rf'TOTAL\s*DIFFERENCE CHARGE.+\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'DIS',
                r'(DIS\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'charge_distance',
                        rf'CHARGE DISTANCE.+\)\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'CTO',
                r'(CTO\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_int_charge',
                        rf'CHARGE\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'CTOi',
                r'(CTO\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_charge_in_sphere',
                        rf'TOTAL\s*CHARGE IN SPHERE.+\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'NECi',
                rf'NEC\d+\:\s*NUCLEAR AND ELECTRONIC CHARGE\s*({re_float})\s*({re_float})',
                dtype=np.dtype(np.float64), repeats=True),
            Quantity(
                'MMINT',
                r'(MMINT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'mmint',
                        rf'MAGNETIC MOMENT IN INTERSTITIAL\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'MMIi',
                r'(MMI\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'mmi',
                        rf'MAGNETIC MOMENT IN SPHERE\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'MMTOT',
                r'(MMTOT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'mmtot',
                        rf' MAGNETIC MOMENT IN CELL\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'ENE',
                r'(ENE\s*\:[\s\S]+?)\n',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_total',
                        rf'TOTAL ENERGY IN Ry\s*\=\s*({re_float})',
                        dtype=np.float64, unit=ureg.rydberg)])),
            Quantity(
                'FORi',
                rf'FOR\d+\:\s*\d+\.ATOM\s*({re_float}\s*{re_float}\s*{re_float}\s*{re_float})',
                repeats=True, dtype=np.dtype(np.float64)),
            Quantity(
                'FGLi',
                rf'FGL\d+\:\s*\d+\.ATOM\s*({re_float}\s*{re_float}\s*{re_float})',
                repeats=True, dtype=np.dtype(np.float64))]

        self._quantities = [
            Quantity(
                'version',
                r'LABEL\d+\:\s*using WIEN2k_(\S+) \(Release ([\d\/]+)\)', flatten=False),
            Quantity(
                'start_date',
                r'LABEL\d+\:\s*on .+ at \w+ (\w+ \d+ \d\d\:\d\d\:\d\d)\s*\w*\s*(\d+)',
                flatten=False),
            Quantity(
                'iteration',
                r'\d+\:\s*(\d+\.\s* ITERATION[\s\S]+?)(?:\:ITE|\Z)',
                repeats=True, sub_parser=TextParser(quantities=iteration_quantities))]


class Wien2kParser:
    def __init__(self):
        # TODO complete list and verify if consistent with current implementation
        # http://www.wien2k.at/reg_user/textbooks/usersguide.pdf
        # implement libxc compatibility
        self._xc_functional_map = {
            '5': ['LDA_X', 'LDA_C_PW'],
            '6': ['HF_X'],
            '11': ['GGA_X_WC', 'GGA_C_PBE'],
            '13': ['GGA_X_PBE', 'GGA_C_PBE'],
            '17': ['GGA_X_PW91'],
            '18': ['HYB_GGA_XC_B3PW91'],
            '19': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            '24': ['GGA_X_B88', 'GGA_C_LYP'],
            '28': ['MGGA_X_TB09', 'LDA_C_PW'],
            '27': ['MGGA_X_TPSS', 'MGGA_C_TPSS'],
            '29': ['MGGA_C_REVTPSS, GGA_C_REGTPSS'],
            '46': ['GGA_X_HTBS'],
            '47': ['HYB_GGA_XC_B3LYP'],
            'XC_LDA': ['LDA_X', 'LDA_C_PW'],
            'XC_PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'XC_WC': ['GGA_X_WC', 'GGA_C_PBE'],
            'XC_PBESOL': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            'XC_B3PW91': ['HYB_GGA_XC_B3PW91'],
            'XC_B3LYP': ['HYB_GGA_XC_B3LYP'],
            'XC_MBJ': ['MGGA_X_TB09', 'LDA_C_PW'],
            'XC_TPSS': ['MGGA_X_TPSS', 'MGGA_C_TPSS'],
            'XC_REVTPSS': ['MGGA_C_REVTPSS, GGA_C_REGTPSS'],
            'XC_MGGA_MS': ['MGGA_X_MS', 'MGGA_C_MS'],
            'XC_MVS': ['MGGA_X_MVS', 'MGGA_C_MVS'],
            'XC_MBEEF': ['MGGA_X_MBEEF', 'GGA_C_PBE_SOL'],
            'XC_SCAN': ['MGGA_X_SCAN', 'MGGA_C_SCAN'],
            'XC_SCANL': ['MGGA_X_SCANL', 'MGGA_C_SCANL'],
            'XC_RSCAN': ['MGGA_X_RSCAN', 'MGGA_C_RSCAN'],
            'XC_R2SCAN': ['MGGA_X_R2SCAN', 'MGGA_C_R2SCAN'],
            'XC_TM': ['MGGA_X_TM', 'MGGA_C_TM'],
            'EC_PW91': ['GGA_X_PW91'],
            'VC_PW91': ['GGA_X_PW91'],
            'EX_B88': ['GGA_X_B88'],
            'VX_B88': ['GGA_X_B88'],
            'EC_LYP': ['GGA_C_LYP'],
            'VC_LYP': ['GGA_C_LYP'],
            'XC_HTBS': ['GGA_X_HTBS'],
            'EX_LDA': ['HF_X'],
            'VX_LDA': ['HF_X']}

        self.out_parser = OutParser()
        self.in0_parser = In0Parser()
        self.in1_parser = In1Parser()
        self.in2_parser = In2Parser()
        self.struct_parser = StructParser()
        self.dos_parser = DosParser()

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger

    def get_wien2k_file(self, ext, multiple=False):
        paths = [p for p in os.listdir(self.maindir) if re.match(r'.*%s$' % ext, p)]
        if not paths:
            return [] if multiple else None
        elif len(paths) == 1:
            path = os.path.join(self.maindir, paths[0])
            return [path] if multiple else path
        else:
            prefix = os.path.basename(self.filepath).rsplit('.', 1)[0]
            paths = [os.path.join(self.maindir, p) for p in paths if p.startswith(prefix)]
            if not paths:
                return [] if multiple else None
            return paths if multiple else paths[0]

    def get_nspin(self):
        return 2 if self.out_parser.get('iteration', [{}])[0].get('VOL', {}).get('spinpolarization') == 'SPINPOLARIZED' else 1

    def get_kpoints(self):
        k_list_file = self.get_wien2k_file('klist')
        if k_list_file is None:
            return None

        kpoints = []
        weights = []
        with open(k_list_file) as f:
            while True:
                line = f.readline()
                if not line or 'END' in line:
                    break
                try:
                    line = np.array(line.split()[:6], dtype=float)
                except Exception:
                    continue
                kpoints.append(line[1:4] / line[4])
                weights.append(line[5])
        weights /= sum(weights)

        return kpoints, weights

    def get_eigenvalues(self):
        nspin = self.get_nspin()
        if nspin == 1:
            files = self.get_wien2k_file(r'energy\_\d+', multiple=True)
            if not files:
                files = [self.get_wien2k_file('energy')]
        else:
            files = []
            for spin in ['up', 'dn']:
                files_spin = self.get_wien2k_file(r'energy%s\_\d+' % spin, multiple=True)
                if not files_spin:
                    files_spin = [self.get_wien2k_file('energy%s' % spin)]
                files.extend(files_spin)
        if None in files:
            return

        re_k = r'\-?\d\.\d+E[\-\+]\d\d'
        re_kpoint = re.compile(rf'\s*({re_k})\s*({re_k})\s*({re_k})\w*\s*(\d+)\s*\d+\s*\d+\s*([\d\.]+)\s*')
        re_eigenvalue = re.compile(r'\s*\d+ +(\-?\d+\.\d+[E\-\+\d]+) *\n*')
        kpoints, eigenvalues, multiplicity, index = [], [], [], []
        for file_i in files:
            with open(file_i) as f:
                while True:
                    line = f.readline()
                    if not line:
                        break
                    if 'energydn' not in file_i:
                        kpoint = re_kpoint.match(line)
                        if kpoint:
                            kpoints.append(kpoint.groups()[:3])
                            index.append(kpoint.group(4))
                            multiplicity.append(kpoint.group(5))
                            continue
                    eigenvalue = re_eigenvalue.match(line)
                    if eigenvalue:
                        eigenvalues.append(eigenvalue.group(1))
        try:
            eigenvalues = np.array(eigenvalues, dtype=np.dtype(np.float64))
            kpoints = np.array(kpoints, dtype=np.dtype(np.float64))
            multiplicity = np.array(multiplicity, dtype=np.dtype(np.float64))
            eigenvalues = np.reshape(eigenvalues, (
                nspin, len(kpoints), len(eigenvalues) // (nspin * len(kpoints))))
            # sort data wrt kpoint index
            index = np.argsort(index)
            kpoints = kpoints[index]
            multiplicity = multiplicity[index]
            for spin in range(nspin):
                eigenvalues[spin] = eigenvalues[spin][index]
            return eigenvalues, kpoints, multiplicity
        except Exception:
            self.logger.error('Error reading eigenvalues.')
            return

    def get_dos(self):
        # dos (projections and total) are printed in dos1, dos2....
        nspin = self.get_nspin()
        if nspin == 1:
            files = self.get_wien2k_file(r'dos\d+', multiple=True)
            files.sort()
        else:
            files = []
            for spin in ['up', 'dn']:
                files_spin = self.get_wien2k_file(r'dos\d+%s' % spin, multiple=True)
                files_spin.sort()
                files.extend(files_spin)

        if not files:
            return

        dos = []
        labels = []
        for file_i in files:
            self.dos_parser.mainfile = file_i
            data = self.dos_parser.get('data')
            if data is None:
                continue
            if not file_i.endswith('dn'):
                labels.extend(self.dos_parser.get('labels', '').split())
            data = np.transpose(data)
            energy = data[0]
            dos.append(data[1:])

        try:
            dos = np.vstack(dos)
            dos = np.reshape(dos, (nspin, len(dos) // nspin, len(dos[0])))
            dos = np.transpose(dos, axes=(1, 0, 2))
            total_dos = []
            partial_dos = []
            for n in range(len(dos)):
                # wien2k may not uniformly print out projections, we get only total projections
                # on the species (independent atoms) denoted by the header N:total
                if labels[n].endswith(':total'):
                    partial_dos.append(dos[n])
                elif labels[n] == 'total-DOS' or labels[n] == 'TOTAL':
                    # TODO determine if total dos is always the last column
                    total_dos = dos[n]
            if len(partial_dos) > 0:
                partial_dos = np.transpose(partial_dos, axes=(1, 0, 2))
            return energy, total_dos, partial_dos

        except Exception:
            self.logger.error('Error reading dos.')
            return

    def parse_scc(self):
        if self.out_parser.get('iteration') is None:
            return

        sec_scc = self.archive.run[0].m_create(Calculation)

        for iteration in self.out_parser.get('iteration'):
            sec_scf = sec_scc.m_create(ScfIteration)
            sec_scf_energy = sec_scf.m_create(Energy)
            for key in iteration.keys():
                if iteration.get(key) is None:
                    continue
                elif key == 'FORi':
                    forces = np.transpose(iteration.get(key))
                    sec_scf.x_wien2k_for = np.transpose(forces[1:4])
                    sec_scf.x_wien2k_for_abs = forces[0]
                elif key == 'FGLi':
                    sec_scf.x_wien2k_for_gl = iteration.get(key)
                elif key == 'MMIi':
                    sec_scf.x_wien2k_mmi = [mm.mmi for mm in iteration.get(key)]
                elif key == 'NECi':
                    charge = np.transpose(iteration.get(key))
                    sec_scf.x_wien2k_nuclear_charge = charge[0]
                    sec_scf.x_wien2k_electronic_charge = charge[1]
                elif key == 'CTOi':
                    charge = [c.tot_charge_in_sphere for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_charge_in_sphere = charge
                elif key == 'DTOi':
                    charge = [c.tot_diff_charge for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_diff_charge = charge
                elif key == 'NTOi':
                    charge = [c.tot_charge_in_sphere_nm for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_charge_in_sphere_nm = charge
                elif key == 'RTOi':
                    density = np.transpose(iteration.get(key))
                    sec_scf.x_wien2k_density_at_nucleus_valence = density[0]
                    sec_scf.x_wien2k_density_at_nucleus_semicore = density[1]
                    sec_scf.x_wien2k_density_at_nucleus_core = density[2]
                    sec_scf.x_wien2k_density_at_nucleus_tot = density[3]
                elif key == 'CHAi':
                    charge = [c.tot_val_charge_cell for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_val_charge_sphere = charge
                elif key == 'POSi':
                    mult = [p.atom_mult for p in iteration.get(key)]
                    sec_scf.x_wien2k_atom_mult = mult
                else:
                    for sub_key, val in iteration.get(key, {}).items():
                        if sub_key.startswith('energy_reference_fermi'):
                            sec_scf_energy.fermi = val
                        elif sub_key.startswith('energy_'):
                            sec_scf_energy.m_add_sub_section(
                                getattr(Energy, sub_key.replace('energy_', '').lower()), EnergyEntry(value=val))
                        else:
                            setattr(sec_scf, 'x_wien2k_%s' % sub_key, val)

        # write final iteration values to scc
        if sec_scf.energy.total is not None:
            sec_scc.energy = Energy(total=EnergyEntry(value=sec_scf.energy.total.value))
            if sec_scf.energy.fermi is not None:
                sec_scc.energy.fermi = sec_scf.energy.fermi

        if sec_scf.x_wien2k_for_gl is not None:
            forces = []
            for n, force in enumerate(sec_scf.x_wien2k_for_gl):
                forces.extend([force] * sec_scf.x_wien2k_atom_mult[n])
            sec_scc.forces = Forces(total=ForcesEntry(value=forces * (ureg.mRy / ureg.bohr)))

        # eigenvalues
        eigenvalues = self.get_eigenvalues()
        if eigenvalues is not None:
            sec_eigenvalues = sec_scc.m_create(BandEnergies)
            sec_eigenvalues.kpoints = eigenvalues[1]
            sec_eigenvalues.kpoints_multiplicities = eigenvalues[2]
            sec_eigenvalues.energies = eigenvalues[0] * ureg.rydberg

        # dos
        dos = self.get_dos()
        if dos is not None:
            # total dos
            if len(dos[1]) > 0:
                sec_dos = sec_scc.m_create(Dos, Calculation.dos_electronic)
                sec_dos.energies = dos[0] * ureg.rydberg
                for spin in range(len(dos[1])):
                    sec_dos_values = sec_dos.m_create(DosValues, Dos.total)
                    sec_dos_values.spin = spin
                    sec_dos_values.value = dos[1][spin] * (1 / ureg.rydberg)

            # projected dos
            if len(dos[2]) > 0:
                labels = [a.atom_name for a in self.struct_parser.get('atom', [])]
                if len(labels) == 0:
                    self.logger.warn('Cannot resolve atom labels.')
                else:
                    for species in range(len(dos[2])):
                        for spin in range(len(dos[2][species])):
                            sec_dos_values = sec_dos.m_create(DosValues, Dos.species_projected)
                            sec_dos_values.atom_label = labels[species]
                            sec_dos_values.spin = spin
                            sec_dos_values.value = dos[2][species][spin] * (1 / ureg.rydberg)

    def parse_system(self):
        sec_system = self.archive.run[0].m_create(System)
        sec_atoms = sec_system.m_create(Atoms)

        self.struct_parser.mainfile = self.get_wien2k_file('struct')
        for key, val in self.struct_parser.get('lattice', {}).items():
            setattr(sec_system, 'x_wien2k_%s' % key, val)

        for atom in self.struct_parser.get('atom', []):
            sec_atom = sec_system.m_create(x_wien2k_section_equiv_atoms)
            for key, val in atom.items():
                setattr(sec_atom, 'x_wien2k_%s' % key, val)

        atoms = self.struct_parser.get_atoms()
        if atoms is None:
            return

        sec_atoms.lattice_vectors = np.array(atoms.get_cell()) * ureg.angstrom
        sec_atoms.positions = atoms.get_positions() * ureg.angstrom
        sec_atoms.labels = atoms.get_chemical_symbols()
        sec_atoms.periodic = atoms.get_pbc()

    def parse_method(self):
        sec_method = self.archive.run[0].m_create(Method)
        sec_dft = sec_method.m_create(DFT)
        sec_electronic = sec_method.m_create(Electronic)

        sec_electronic.n_spin_channels = self.get_nspin()

        # read functional settings from in0 file
        self.in0_parser.mainfile = self.get_wien2k_file('in0')

        # better to read it from scf?
        xc_functional = self.in0_parser.get('xc_functional', None)
        xc_functional = xc_functional if isinstance(xc_functional, list) else [xc_functional]
        sec_xc_functional = sec_dft.m_create(XCFunctional)
        for name in xc_functional:
            functionals = self._xc_functional_map.get(name)
            if functionals is None:
                self.logger.warn('Cannot resolve XC functional.')
                continue
            for functional in functionals:
                if '_X_' in functional or functional.endswith('_X'):
                    sec_xc_functional.exchange.append(Functional(name=functional))
                elif '_C_' in functional or functional.endswith('_C'):
                    sec_xc_functional.correlation.append(Functional(name=functional))
                elif 'HYB' in functional:
                    sec_xc_functional.hybrid.append(Functional(name=functional))
                else:
                    sec_xc_functional.contributions.append(Functional(name=functional))

        fft = self.in0_parser.get('fft')
        if fft is not None:
            sec_method.x_wien2k_ifft = fft[:3]
            sec_method.x_wien2k_ifft_factor = fft[3]

        # read cut off settings from in1
        in1_file = self.get_wien2k_file('in1')
        if in1_file is None:
            in1_file = self.get_wien2k_file('in1c')
        self.in1_parser.mainfile = in1_file
        for key, val in self.in1_parser.items():
            if val is not None:
                setattr(sec_method, 'x_wien2k_%s' % key, val)

        # read integration data from in2 file
        in2_file = self.get_wien2k_file('1n2')
        if in2_file is None:
            in2_file = self.get_wien2k_file('in2c')
        self.in2_parser.mainfile = in2_file

        for key in ['gmax', 'switch']:
            val = self.in2_parser.get(key)
            if val is not None:
                setattr(sec_method, 'x_wien2k_in2_%s' % key, val)

        emin_keys = ['emin', 'ne', 'espermin', 'esper0']
        for n, val in enumerate(self.in2_parser.get('emin', [])):
            if n < 4:
                setattr(sec_method, 'x_wien2k_in2_%s' % emin_keys[n], val)

        smearing, width = self.in2_parser.get('smearing', [None, None])
        if smearing is not None:
            sec_smearing = sec_electronic.m_create(Smearing)
            if smearing.startswith('GAUSS'):
                smearing = 'gaussian'
            elif smearing.startswith('TEMP'):
                smearing = 'fermi'
            elif smearing.startswith('TETRA'):
                smearing = 'tetrahedra'
            sec_smearing.kind = smearing
            sec_smearing.width = (float(width) * ureg.rydberg).to('joule').magnitude

        # read kpoints from klist
        kpoints = self.get_kpoints()
        if kpoints is not None:
            sec_k_mesh = sec_method.m_create(KMesh)
            sec_k_mesh.points = kpoints[0]
            sec_k_mesh.weights = kpoints[1]

        # basis
        sec_method.basis_set.append(BasisSet(type='(L)APW+lo'))

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)

        sec_run.program = Program(name='WIEN2k', version=self.out_parser.get('version', ''))
        start_date = self.out_parser.get('start_date')
        if start_date is not None:
            # TODO resolve proper timezone
            dt = datetime.strptime(start_date, '%b %d %H:%M:%S %Y') - datetime.utcfromtimestamp(0)
            sec_run.time_run = TimeRun(date_start=dt.total_seconds())

        # TODO implement geometry optimization
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.type = 'single_point'

        self.parse_method()

        self.parse_system()

        self.parse_scc()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Definitions that are used in the documentation via mkdocs-macro-plugin.
'''

from nomad.app.v1.models import (
    query_documentation,
    owner_documentation)
from nomad.app.v1.routers.entries import archive_required_documentation
from nomad import config, utils

doc_snippets = {
    'query': query_documentation,
    'owner': owner_documentation,
    'archive-required': archive_required_documentation
}


def define_env(env):
    @env.macro
    def nomad_url():  # pylint: disable=unused-variable
        return config.api_url()

    @env.macro
    def doc_snippet(key):  # pylint: disable=unused-variable
        return doc_snippets[key]

    @env.macro
    def metainfo_data():  # pylint: disable=unused-variable
        return utils.strip('''
            You can browse the [NOMAD metainfo schema](../gui/analyze/metainfo)
            or the archive of each entry (e.g. [a VASP example](../gui/search/entries/entry/id/d5OYC0SJTDevHMPk7YHd4A/-7j8ojKkna2NLXdytv_OjV4zsBXw/archive))
            in the web-interface.''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_gamess_section_atom_forces(MSection):
    '''
    section that contains Cartesian forces of the system for a given geometry
    '''

    m_def = Section(validate=False)

    x_gamess_atom_x_force = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        -
        ''')

    x_gamess_atom_y_force = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        -
        ''')

    x_gamess_atom_z_force = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        -
        ''')


class x_gamess_section_cis(MSection):
    '''
    Configuration interaction singles excitation energies and oscillator strengths.
    '''

    m_def = Section(validate=False)

    x_gamess_cis_excitation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the excitation energies for configuration interaction singles excited
        states.
        ''')

    x_gamess_cis_oscillator_strength = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the oscillator strengths for configuration interaction singles excited
        states.
        ''')


class x_gamess_section_ci(MSection):
    '''
    Configuration interaction energies.
    '''

    m_def = Section(validate=False)


class x_gamess_section_coupled_cluster(MSection):
    '''
    Coupled cluster energies.
    '''

    m_def = Section(validate=False)


class x_gamess_section_elstruc_method(MSection):
    '''
    Section containing electronic structure method.
    '''

    m_def = Section(validate=False)

    x_gamess_electronic_structure_method = Quantity(
        type=str,
        shape=[],
        description='''
        Name of electronic structure method.
        ''')


class x_gamess_section_excited_states(MSection):
    '''
    Time-dependent DFT and configuration interaction singles results.
    '''

    m_def = Section(validate=False)

    x_gamess_section_cis = SubSection(
        sub_section=SectionProxy('x_gamess_section_cis'),
        repeats=True)

    x_gamess_section_tddft = SubSection(
        sub_section=SectionProxy('x_gamess_section_tddft'),
        repeats=True)


class x_gamess_section_frequencies(MSection):
    '''
    section for the values of the frequencies, reduced masses and normal mode vectors
    '''

    m_def = Section(validate=False)

    x_gamess_frequencies = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_frequencies'],
        description='''
        values of frequencies, in cm-1
        ''')

    x_gamess_frequency_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of frequencies, in cm-1
        ''')

    x_gamess_red_masses = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_frequencies'],
        description='''
        values of normal mode reduced masses
        ''')

    x_gamess_reduced_masses = Quantity(
        type=str,
        shape=['number_of_reduced_masses_rows'],
        description='''
        values of normal mode reduced masses
        ''')


class x_gamess_section_geometry_optimization_info(MSection):
    '''
    Specifies whether a geometry optimization is converged.
    '''

    m_def = Section(validate=False)

    x_gamess_geometry_optimization_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies whether a geometry optimization is converged.
        ''')


class x_gamess_section_geometry(MSection):
    '''
    section that contains Cartesian coordinates of the system for a given geometry
    '''

    m_def = Section(validate=False)

    x_gamess_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Initial positions of all the atoms, in Cartesian coordinates.
        ''')

    x_gamess_atom_x_coord = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x coordinate for the atoms
        ''')

    x_gamess_atom_y_coord = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y coordinate for the atoms
        ''')

    x_gamess_atom_z_coord = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z coordinate for the atoms
        ''')


class x_gamess_section_mcscf(MSection):
    '''
    Multiconfigurational SCF energies.
    '''

    m_def = Section(validate=False)

    x_gamess_energy_mcscf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the MCSCF total energy, normally CASSCF, during the iterations.
        ''')

    x_gamess_mcscf_active_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of MCSCF active electrons in the calculation.
        ''')

    x_gamess_mcscf_active_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of MCSCF active orbitals in the calculation.
        ''')

    x_gamess_mcscf_inactive_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of MCSCF inactive orbitals in the calculation.
        ''')


class x_gamess_section_moller_plesset(MSection):
    '''
    Perturbative Moller-Plesset energies.
    '''

    m_def = Section(validate=False)


class x_gamess_section_mrpt2(MSection):
    '''
    Multiference multiconfigurational energies at second order of perturbation theory.
    '''

    m_def = Section(validate=False)

    x_gamess_mrpt2_active_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of active electrons in MRPT2 calculation.
        ''')

    x_gamess_mrpt2_active_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of active orbitals in MRPT2 calculation.
        ''')

    x_gamess_mrpt2_doubly_occupied_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of doubly occupied orbitals in MRPT2 calculation.
        ''')

    x_gamess_mrpt2_external_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of external orbitals in MRPT2 calculation.
        ''')

    x_gamess_mrpt2_frozen_core_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of frozen core orbitals in MRPT2 calculation.
        ''')

    x_gamess_mrpt2_frozen_virtual_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of frozen virtual orbitals in MRPT2 calculation.
        ''')

    x_gamess_mrpt2_method_type = Quantity(
        type=str,
        shape=[],
        description='''
        Determinant (MRPT2) or CSF (MC-QDPT) method for second-order perturbation theory
        calculation
        ''')


class x_gamess_section_scf_hf_method(MSection):
    '''
    Section containing type of SCF method employed (RHF,UHF,ROHF or GVB).
    '''

    m_def = Section(validate=False)

    x_gamess_scf_hf_method = Quantity(
        type=str,
        shape=[],
        description='''
        Type of SCF method employed.
        ''')


class x_gamess_section_tddft(MSection):
    '''
    Time-dependent DFT excitation energies and oscillator strengths.
    '''

    m_def = Section(validate=False)

    x_gamess_tddft_excitation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the excitation energies for time-dependent DFT excited states.
        ''')

    x_gamess_tddft_oscillator_strength = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the oscillator strengths for time-dependent DFT excited states.
        ''')


class BandEnergies(simulation.calculation.BandEnergies):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_alpha_eigenvalues_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of eigenenergies for alpha MOs
        ''')

    x_gamess_beta_eigenvalues_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of eigenenergies for occupied beta MOs
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_atom_positions_initial = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Initial positions of all the atoms, in Cartesian coordinates.
        ''')

    x_gamess_atom_x_coord_initial = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x coordinate for the atoms of the initial geometry
        ''')

    x_gamess_atom_y_coord_initial = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y coordinate for the atoms of the initial geometry
        ''')

    x_gamess_atom_z_coord_initial = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z coordinate for the atoms of the initial geometry
        ''')

    x_gamess_number_of_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of electrons for system
        ''')

    x_gamess_atomic_number = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic number for atoms
        ''')

    x_gamess_memory = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total memory for GAMESS job
        ''')

    x_gamess_spin_target_multiplicity = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Target (user-imposed) value of the spin multiplicity $M=2S+1$, where $S$ is the
        total spin. It is an integer value. This value is not necessarly the value
        obtained at the end of the calculation. See spin_S2 for the converged value of the
        spin moment.
        ''')

    x_gamess_total_charge = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total charge of the system.
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_basis_set_diffsp = Quantity(
        type=str,
        shape=[],
        description='''
        Include a set of SP diffuse functions on heavy atoms or not
        ''')

    x_gamess_basis_set_diffs = Quantity(
        type=str,
        shape=[],
        description='''
        Incluse a set of S diffuse functions on light atoms or not
        ''')

    x_gamess_basis_set_gbasis = Quantity(
        type=str,
        shape=[],
        description='''
        Gaussian basis set main name
        ''')

    x_gamess_basis_set_igauss = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of main Gaussians
        ''')

    x_gamess_basis_set_ndfunc = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of polarization d function sets on heavy atoms
        ''')

    x_gamess_basis_set_nffunc = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of polarization f function sets on heavy atoms
        ''')

    x_gamess_basis_set_npfunc = Quantity(
        type=str,
        shape=[],
        description='''
        Number of polarization p function sets on light atoms
        ''')

    x_gamess_basis_set_polar = Quantity(
        type=str,
        shape=[],
        description='''
        Exponents of polarization functions
        ''')

    x_gamess_cctype = Quantity(
        type=str,
        shape=[],
        description='''
        Type of coupled cluster method employed
        ''')

    x_gamess_cistep = Quantity(
        type=str,
        shape=[],
        description='''
        Determinant or CSF method for multiconfigurational SCF calculation
        ''')

    x_gamess_citype = Quantity(
        type=str,
        shape=[],
        description='''
        Type of configuration interaction method employed
        ''')

    x_gamess_comp_method = Quantity(
        type=str,
        shape=[],
        description='''
        Control if the G3MP2 composite method has been defined
        ''')

    x_gamess_mcscf_casscf = Quantity(
        type=str,
        shape=[],
        description='''
        This indicates whether the multiconfigurational SCF calculation is a complete
        active space SCF calculation or not
        ''')

    x_gamess_method = Quantity(
        type=str,
        shape=[],
        description='''
        String identifying in an unique way the WF method used for the final
        wavefunctions.
        ''')

    x_gamess_mplevel = Quantity(
        type=str,
        shape=[],
        description='''
        Level of second-orden Moller-Plesset perturbation theory
        ''')

    x_gamess_pptype = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the pseudopotential employed
        ''')

    x_gamess_relatmethod = Quantity(
        type=str,
        shape=[],
        description='''
        Type of relativistic method employed
        ''')

    x_gamess_scf_method = Quantity(
        type=str,
        shape=[],
        description='''
        String identifying in an unique way the SCF method used in the calculation.
        ''')

    x_gamess_scf_type = Quantity(
        type=str,
        shape=[],
        description='''
        String identifying in an unique way the SCF method used in the calculation.
        ''')

    x_gamess_tddfttype = Quantity(
        type=str,
        shape=[],
        description='''
        Type of time-dependent DFT calculation
        ''')

    x_gamess_vbtype = Quantity(
        type=str,
        shape=[],
        description='''
        Type of valence bond method employed
        ''')

    x_gamess_xc = Quantity(
        type=str,
        shape=[],
        description='''
        String identifying in an unique way the XC method used for the final
        wavefunctions.
        ''')

    x_gamess_control_options = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_gamess_system_options = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_gamess_basis_options = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_gamess_section_elstruc_method = SubSection(
        sub_section=SectionProxy('x_gamess_section_elstruc_method'),
        repeats=True)

    x_gamess_section_scf_hf_method = SubSection(
        sub_section=SectionProxy('x_gamess_section_scf_hf_method'),
        repeats=True)


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_gamess_program_implementation = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_gamess_section_geometry_optimization_info = SubSection(
        sub_section=SectionProxy('x_gamess_section_geometry_optimization_info'),
        repeats=True)


class ScfIteration(simulation.calculation.ScfIteration):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_energy_scf = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Final value of the total electronic energy calculated with the method described in
        XC_method.
        ''')

    x_gamess_energy_total_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the total electronic energy calculated with the method described in
        XC_method during each self-consistent field (SCF) iteration.
        ''')


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_section_atom_forces = SubSection(
        sub_section=SectionProxy('x_gamess_section_atom_forces'),
        repeats=True)

    x_gamess_section_ci = SubSection(
        sub_section=SectionProxy('x_gamess_section_ci'),
        repeats=True)

    x_gamess_section_coupled_cluster = SubSection(
        sub_section=SectionProxy('x_gamess_section_coupled_cluster'),
        repeats=True)

    x_gamess_section_excited_states = SubSection(
        sub_section=SectionProxy('x_gamess_section_excited_states'),
        repeats=True)

    x_gamess_section_frequencies = SubSection(
        sub_section=SectionProxy('x_gamess_section_frequencies'),
        repeats=True)

    x_gamess_section_geometry = SubSection(
        sub_section=SectionProxy('x_gamess_section_geometry'),
        repeats=True)

    x_gamess_section_mcscf = SubSection(
        sub_section=SectionProxy('x_gamess_section_mcscf'),
        repeats=True)

    x_gamess_section_moller_plesset = SubSection(
        sub_section=SectionProxy('x_gamess_section_moller_plesset'),
        repeats=True)

    x_gamess_section_mrpt2 = SubSection(
        sub_section=SectionProxy('x_gamess_section_mrpt2'),
        repeats=True)


class Energy(simulation.calculation.Energy):

    m_def = Section(validate=False, extends_base_section=True)

    x_gamess_virial_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')
    x_gamess_wavefunction_normalization = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_gamess_one_electron = SubSection(
        sub_section=simulation.calculation.EnergyEntry.m_def,
        description='''
        ''')

    x_gamess_two_electron = SubSection(
        sub_section=simulation.calculation.EnergyEntry.m_def,
        description='''
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import

from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_fleur_header(MSection):
    '''
    header (labels) of fleur.
    '''

    m_def = Section(validate=False)

    x_fleur_version = Quantity(
        type=str,
        shape=[],
        description='''
        Version of Fleur
        ''')

    x_fleur_precision = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_fleur_with_inversion = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_fleur_with_soc = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_fleur_additional_flags = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class x_fleur_section_equiv_atoms(MSection):
    '''
    section containing a class of equivalent atoms
    '''

    m_def = Section(validate=False)

    x_fleur_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        name of atom, labelling non-equvalent atoms
        ''')

    x_fleur_atom_pos_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        position of atom x
        ''')

    x_fleur_atom_pos_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        position of atom y
        ''')

    x_fleur_atom_pos_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        position of atom z
        ''')

    x_fleur_atom_coord_scale = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        scales coordinates by 1/scale. If film=T, scales only x&y coordinates, if film=F
        also z
        ''')

    x_fleur_atomic_number_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic number Z
        ''')

    x_fleur_nr_equiv_atoms_in_this_atom_type = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number_equiv_atoms_in_this_atom_type
        ''')


class x_fleur_section_XC(MSection):
    '''
    exchange-correlation potential
    '''

    m_def = Section(validate=False)

    x_fleur_exch_pot = Quantity(
        type=str,
        shape=[],
        description='''
        exchange-correlation potential, in out
        ''')

    x_fleur_xc_correction = Quantity(
        type=str,
        shape=[],
        description='''
        informaion on relativistic correction for the exchange-correlation potential, in
        out
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_fleur_header = SubSection(
        sub_section=SectionProxy('x_fleur_header'),
        repeats=True)


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_fleur_lattice_vector_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of vector of unit cell
        ''')

    x_fleur_lattice_vector_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of vector of unit cell
        ''')

    x_fleur_lattice_vector_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of vector of unit cell
        ''')

    x_fleur_rec_lattice_vector_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of reciprocal lattice vector
        ''')

    x_fleur_rec_lattice_vector_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of reciprocal lattice vector
        ''')

    x_fleur_rec_lattice_vector_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of reciprocal lattice vector
        ''')

    x_fleur_space_group = Quantity(
        type=str,
        shape=[],
        description='''
        space group
        ''')

    x_fleur_name_of_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        name of atom type
        ''')

    x_fleur_system_nameIn = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system given in the inp file
        ''')

    x_fleur_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system
        ''')

    x_fleur_total_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        total number of atoms
        ''')

    x_fleur_nr_of_atom_types = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of atom types
        ''')

    x_fleur_nuclear_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        nuclear number
        ''')

    x_fleur_number_of_core_levels = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_fleur_number_of_core_levels
        ''')

    x_fleur_lexpansion_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        l-expansion cutoff
        ''')

    x_fleur_mt_gridpoints = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        muffin-tin gridpoints
        ''')

    x_fleur_mt_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        muffin-tin radius
        ''')

    x_fleur_logarythmic_increment = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        logarythmic increment
        ''')

    x_fleur_k_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Kmax is the plane wave cut-off
        ''')

    x_fleur_G_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Gmax
        ''')

    x_fleur_tot_nucl_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total nuclear charge
        ''')

    x_fleur_tot_elec_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total electronic charge
        ''')

    x_fleur_unit_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='bohr ** 3',
        description='''
        unit cell volume
        ''')

    x_fleur_unit_cell_volume_omega = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell volume omega tilda
        ''')

    x_fleur_vol_interstitial = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        volume of interstitial region
        ''')

    x_fleur_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_fleur_section_equiv_atoms = SubSection(
        sub_section=SectionProxy('x_fleur_section_equiv_atoms'),
        repeats=True)


class ScfIteration(simulation.calculation.ScfIteration):

    m_def = Section(validate=False, extends_base_section=True)

    x_fleur_tot_for_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, X
        ''')

    x_fleur_tot_for_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, Y
        ''')

    x_fleur_tot_for_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, Z
        ''')

    x_fleur_tot_for_fx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, FX_TOT
        ''')

    x_fleur_tot_for_fy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, FY_TOT
        ''')

    x_fleur_tot_for_fz = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, FZ_TOT
        ''')

    x_fleur_iteration_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        scf iteration number
        ''')

    x_fleur_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        energy total
        ''')

    x_fleur_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        free energy
        ''')

    x_fleur_entropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        (tkb*entropy) TS
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_fleur_nkptd = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of all the k-points
        ''')

    x_fleur_k_point_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x component of vector of k point
        ''')

    x_fleur_k_point_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        y component of vector of k point
        ''')

    x_fleur_k_point_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        z component of vector of k point
        ''')

    x_fleur_k_point_weight = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        weights of k point
        ''')

    x_fleur_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        The Brillouin zone integration mode. It can be one of hist - Use the histogram
        mode, this is the default; gauss - Use Gaussian smearing, tria - Use the
        tetrahedron method
        ''')

    x_fleur_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        specifies the width of the broadening, smearing for calculation of fermi-energy &
        weights. The Fermi smearing can be parametrized by this energy
        ''')

    x_fleur_nr_of_valence_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The number of electrons to be represented within the valence electron framework
        ''')

    x_fleur_smearing_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        Fermi smearing temperature set in Kelvin
        ''')

    x_fleur_input_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_fleur_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_fleur_eigenvalues_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_fleur_section_XC = SubSection(
        sub_section=SectionProxy('x_fleur_section_XC'),
        repeats=True)


class XCFunctional(simulation.method.XCFunctional):

    m_def = Section(validate=False, extends_base_section=True)

    x_fleur_xc_correction = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class BaseCalculation(simulation.calculation.BaseCalculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_fleur_n_occupied_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation
from nomad.datamodel.metainfo import workflow


m_package = Package()


class x_dl_poly_section_md_molecule_type(MSection):
    '''
    Section to store molecule type information
    '''

    m_def = Section(validate=False)

    x_dl_poly_md_molecule_type_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Molecule type id
        ''')

    x_dl_poly_md_molecule_type_name = Quantity(
        type=str,
        shape=[],
        description='''
        Molecule type name
        ''')


class x_dl_poly_section_md_topology(MSection):
    '''
    Section modelling the MD topology
    '''

    m_def = Section(validate=False)

    x_dl_poly_md_molecular_types = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of molecular types in topology
        ''')

    x_dl_poly_section_md_molecule_type = SubSection(
        sub_section=SectionProxy('x_dl_poly_section_md_molecule_type'),
        repeats=True)


class MolecularDynamics(workflow.MolecularDynamics):

    m_def = Section(validate=False, extends_base_section=True)

    x_dl_poly_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''')

    x_dl_poly_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''')

    x_dl_poly_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''')

    x_dl_poly_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''')

    x_dl_poly_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''')

    x_dl_poly_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''')

    x_dl_poly_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_dl_poly_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date
        ''')

    x_dl_poly_system_description = Quantity(
        type=str,
        shape=[],
        description='''
        Simulation run title
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_dl_poly_section_md_topology = SubSection(
        sub_section=SectionProxy('x_dl_poly_section_md_topology'),
        repeats=True)


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_dl_poly_step_number_equilibration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MD equilibration step number
        ''')

    x_dl_poly_step_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MD total step number
        ''')

    x_dl_poly_thermostat_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Thermostat coupling temperature
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

''' All generic entry metadata and related classes. '''

from typing import List, Any
from enum import Enum
from cachetools import cached, TTLCache
from elasticsearch_dsl import analyzer, tokenizer

from nomad import metainfo, config
from nomad.metainfo.mongoengine_extension import Mongo, MongoDocument
from nomad.datamodel.metainfo.common import FastAccess
from nomad.metainfo.pydantic_extension import PydanticModel
from nomad.metainfo.elasticsearch_extension import Elasticsearch, material_entry_type, entry_type as es_entry_type

# This is usually defined automatically when the first metainfo definition is evaluated, but
# due to the next imports requireing the m_package already, this would be too late.
m_package = metainfo.Package()

from .results import Results  # noqa
from .data import EntryData  # noqa
from .optimade import OptimadeEntry  # noqa
from .metainfo.simulation.run import Run  # noqa
from .metainfo.workflow import Workflow  # noqa
from .metainfo.measurements import Measurement  # noqa
from .metainfo.tabulartree import TabularTree  # noqa


class AuthLevel(int, Enum):
    '''
    Used to decorate fields with the authorization level required to edit them (using `a_auth_level`).
    * `none`: No authorization required
    * `coauthor`: You must be at least a coauthor of the upload to edit the field.
    * `main_author`: You must be the main author of the upload to edit the field.
    * `admin`: You must be admin to edit the field.
    '''
    none = 0
    coauthor = 1
    main_author = 2
    admin = 3


path_analyzer = analyzer(
    'path_analyzer',
    tokenizer=tokenizer('path_tokenizer', 'pattern', pattern='/'))


def PathSearch():
    return [
        Elasticsearch(_es_field='keyword'),
        Elasticsearch(
            mapping=dict(type='text', analyzer=path_analyzer.to_dict()),
            field='path', _es_field='')]


quantity_analyzer = analyzer(
    'quantity_analyzer',
    tokenizer=tokenizer('quantity_tokenizer', 'pattern', pattern='.'))


def QuantitySearch():
    return [
        Elasticsearch(material_entry_type, _es_field='keyword'),
        Elasticsearch(
            material_entry_type,
            mapping=dict(type='text', analyzer=path_analyzer.to_dict()),
            field='path', _es_field='')]


class Author(metainfo.MSection):
    ''' A person that is author of data in NOMAD or references by NOMAD. '''
    name = metainfo.Quantity(
        type=str,
        derived=lambda user: ('%s %s' % (user.first_name, user.last_name)).strip(),
        a_elasticsearch=[
            Elasticsearch(material_entry_type, _es_field='keyword'),
            Elasticsearch(material_entry_type, mapping='text', field='text', _es_field=''),
            Elasticsearch(suggestion="default")
        ])

    first_name = metainfo.Quantity(type=metainfo.Capitalized)
    last_name = metainfo.Quantity(type=metainfo.Capitalized)
    email = metainfo.Quantity(type=str)

    affiliation = metainfo.Quantity(type=str)
    affiliation_address = metainfo.Quantity(type=str)


class User(Author):
    ''' A NOMAD user.

    Typically a NOMAD user has a NOMAD account. The user related data is managed by
    NOMAD keycloak user-management system. Users are used to denote authors,
    reviewers, and owners of datasets.

    Args:
        user_id: The unique, persistent keycloak UUID
        username: The unique, persistent, user chosen username
        first_name: The users first name (including all other given names)
        last_name: The users last name
        affiliation: The name of the company and institutes the user identifies with
        affiliation_address: The address of the given affiliation
        created: The time the account was created
        repo_user_id: The id that was used to identify this user in the NOMAD CoE Repository
        is_admin: Bool that indicated, iff the user the use admin user
    '''

    m_def = metainfo.Section(a_pydantic=PydanticModel())

    user_id = metainfo.Quantity(
        type=str,
        a_elasticsearch=Elasticsearch(material_entry_type))

    username = metainfo.Quantity(type=str)

    created = metainfo.Quantity(type=metainfo.Datetime)

    repo_user_id = metainfo.Quantity(
        type=str,
        description='Optional, legacy user id from the old NOMAD CoE repository.')

    is_admin = metainfo.Quantity(
        type=bool, derived=lambda user: user.user_id == config.services.admin_user_id)

    is_oasis_admin = metainfo.Quantity(type=bool, default=False)

    @staticmethod
    @cached(cache=TTLCache(maxsize=2048, ttl=24 * 3600))
    def get(*args, **kwargs) -> 'User':
        from nomad import infrastructure
        return infrastructure.keycloak.get_user(*args, **kwargs)  # type: ignore

    def full_user(self) -> 'User':
        ''' Returns a User object with all attributes loaded from the user management system. '''
        from nomad import infrastructure
        assert self.user_id is not None
        return infrastructure.keycloak.get_user(user_id=self.user_id)  # type: ignore


class UserReference(metainfo.Reference):
    '''
    Special metainfo reference type that allows to use user_ids as values. It automatically
    resolves user_ids to User objects. This is done lazily on getting the value.
    '''

    def __init__(self):
        super().__init__(User.m_def)

    def resolve(self, proxy: metainfo.MProxy) -> metainfo.MSection:
        return User.get(user_id=proxy.m_proxy_value)

    def serialize(self, section: metainfo.MSection, quantity_def: metainfo.Quantity, value: Any) -> Any:
        return value.user_id


user_reference = UserReference()


class AuthorReference(metainfo.Reference):
    '''
    Special metainfo reference type that allows to use either user_ids or direct author
    information as values. It automatically resolves user_ids to User objects and author
    data into Author objects.
    '''

    def __init__(self):
        super().__init__(Author.m_def)

    def resolve(self, proxy: metainfo.MProxy) -> metainfo.MSection:
        proxy_value = proxy.m_proxy_value
        if isinstance(proxy_value, str):
            return User.get(user_id=proxy.m_proxy_value)
        elif isinstance(proxy_value, dict):
            return Author.m_from_dict(proxy_value)
        else:
            raise metainfo.MetainfoReferenceError()

    def serialize(self, section: metainfo.MSection, quantity_def: metainfo.Quantity, value: Any) -> Any:
        if isinstance(value, User):
            return value.user_id
        elif isinstance(value, Author):
            return value.m_to_dict()
        else:
            raise metainfo.MetainfoReferenceError()


author_reference = AuthorReference()


class Dataset(metainfo.MSection):
    ''' A Dataset is attached to one or many entries to form a set of data.

    Args:
        dataset_id: The unique identifier for this dataset as a string. It should be
            a randomly generated UUID, similar to other nomad ids.
        dataset_name: The human readable name of the dataset as string. The dataset name must be
            unique for the user.
        user_id: The unique user_id of the owner and creator of this dataset. The owner
            must not change after creation.
        doi: The optional Document Object Identifier (DOI) associated with this dataset.
            Nomad can register DOIs that link back to the respective representation of
            the dataset in the nomad UI. This quantity holds the string representation of
            this DOI. There is only one per dataset. The DOI is just the DOI name, not its
            full URL, e.g. "10.17172/nomad/2019.10.29-1".
        pid: The original NOMAD CoE Repository dataset PID. Old DOIs still reference
            datasets based on this id. Is not used for new datasets.
        dataset_create_time: The date when the dataset was first created.
        dataset_modified_time: The date when the dataset was last modified. An owned dataset
            can only be extended after a DOI was assigned. A foreign dataset cannot be changed
            once a DOI was assigned.
        dataset_type: The type determined if a dataset is owned, i.e. was created by
            the authors of the contained entries; or if a dataset is foreign,
            i.e. it was created by someone not necessarily related to the entries.
    '''
    m_def = metainfo.Section(a_mongo=MongoDocument(), a_pydantic=PydanticModel())

    dataset_id = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(primary_key=True),
        a_elasticsearch=Elasticsearch(material_entry_type))
    dataset_name = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True),
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion="default"),
        ])
    user_id = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True))
    doi = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True),
        a_elasticsearch=Elasticsearch(material_entry_type))
    pid = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True))
    dataset_create_time = metainfo.Quantity(
        type=metainfo.Datetime,
        a_mongo=Mongo(index=True),
        a_elasticsearch=Elasticsearch())
    dataset_modified_time = metainfo.Quantity(
        type=metainfo.Datetime,
        a_mongo=Mongo(index=True),
        a_elasticsearch=Elasticsearch())
    dataset_type = metainfo.Quantity(
        type=metainfo.MEnum('owned', 'foreign'),
        a_mongo=Mongo(index=True),
        a_elasticsearch=Elasticsearch())
    query = metainfo.Quantity(
        type=metainfo.JSON, a_mongo=Mongo())
    entries = metainfo.Quantity(
        type=str, shape=['*'], a_mongo=Mongo())


class DatasetReference(metainfo.Reference):
    '''
    Special metainfo reference type that allows to use dataset_ids as values. It automatically
    resolves dataset_ids to Dataset objects. This is done lazily on getting the value.
    '''

    def __init__(self):
        super().__init__(Dataset.m_def)

    def resolve(self, proxy: metainfo.MProxy) -> metainfo.MSection:
        return Dataset.m_def.a_mongo.get(dataset_id=proxy.m_proxy_value)

    def serialize(self, section: metainfo.MSection, quantity_def: metainfo.Quantity, value: Any) -> Any:
        if isinstance(value, metainfo.MProxy):
            return value.m_proxy_value
        else:
            return value.dataset_id


dataset_reference = DatasetReference()


class EditableUserMetadata(metainfo.MCategory):
    ''' NOMAD entry metadata quantities that can be edited by the user before or after publish. '''
    pass


class MongoUploadMetadata(metainfo.MCategory):
    ''' The field is defined on the Upload mongo document. '''
    pass


class MongoEntryMetadata(metainfo.MCategory):
    ''' The field is defined on the Entry mongo document. '''
    pass


class MongoSystemMetadata(metainfo.MCategory):
    '''
    The field is managed directly by the system/process (or derived from data managed by the
    system/process), and should never be updated from an :class:`EntryMetadata` object.
    '''
    pass


class DomainMetadata(metainfo.MCategory):
    ''' NOMAD entry quantities that are determined by the uploaded data. '''
    pass


def derive_origin(entry: 'EntryMetadata') -> str:
    if entry.external_db is not None:
        return str(entry.external_db)

    if entry.main_author:
        return entry.main_author.name

    return None


def derive_authors(entry: 'EntryMetadata') -> List[User]:
    if entry.external_db == 'EELS Data Base':
        return list(entry.entry_coauthors)

    authors: List[User] = [entry.main_author]
    if entry.coauthors:
        authors.extend(entry.coauthors)
    if entry.entry_coauthors:
        authors.extend(entry.entry_coauthors)
    return authors


class EntryMetadata(metainfo.MSection):
    '''
    Attributes:
        upload_id: The id of the upload (random UUID).
        upload_name: The user provided upload name.
        upload_create_time: The time that the upload was created
        entry_id: The unique mainfile based entry id.
        entry_hash: The raw file content based checksum/hash of this entry.
        entry_create_time: The time that the entry was created
        last_edit_time: The date and time the user metadata was last edited.
        parser_name: The NOMAD parser used for the last processing.
        mainfile: The path to the mainfile from the root directory of the uploaded files.
        mainfile_key: Key used to differentiate between different *child entries* of an entry.
            For parent entries and entries that do not have any children, the value should
            be empty.
        files: A list of all files, relative to upload.
        pid: The unique, sequentially enumerated, integer PID that was used in the legacy
            NOMAD CoE. It allows to resolve URLs of the old NOMAD CoE Repository.
        raw_id: The code specific identifier extracted from the entry's raw files by the parser,
            if supported.
        external_id: A user provided external id. Usually the id for an entry in an external
            database where the data was imported from.
        published: Indicates if the entry is published.
        publish_time: The time when the upload was published.
        with_embargo: Indicated if this entry is under an embargo. Entries with embargo are
            only visible to the main author, the upload coauthors, and the upload reviewers.
        license: A short license description (e.g. CC BY 4.0), that refers to the
            license of this entry.
        processed: Boolean indicating if this entry was successfully processed and archive
            data and entry metadata is available.
        last_processing_time: The date and time of the last processing.
        processing_errors: Errors that occured during processing.
        nomad_version: A string that describes the version of the nomad software that was
            used to do the last successful processing.
        nomad_commit: The NOMAD commit used for the last processing.
        comment: An arbitrary string with user provided information about the entry.
        references: A list of URLs for resources that are related to the entry.
        external_db: The repository or external database where the original data resides.
        origin: A short human readable description of the entries origin. Usually it is the
            handle of an external database/repository or the name of the main author.
        main_author: Id of the main author of this entry.
        coauthors: A user provided list of co-authors for the whole upload. These can view
            and edit the upload when in staging, and view it also if it is embargoed.
        entry_coauthors: Ids of all co-authors (excl. the main author and upload coauthors)
            specified on the entry level, rather than on the upload level. They are shown
            as authors of this entry alongside its main author and upload coauthors.
        reviewers: Ids of users who can review the upload which this entry belongs to. Like
            the main author and the upload coauthors, reviewers can find, see, and download
            all data from the upload and all its entries, even if it is in staging or has
            an embargo.
        datasets: Ids of all datasets that this entry appears in
    '''
    m_def = metainfo.Section(label='Metadata')

    upload_id = metainfo.Quantity(
        type=str, categories=[MongoUploadMetadata],
        description='The persistent and globally unique identifier for the upload of the entry',
        a_elasticsearch=Elasticsearch(material_entry_type, metrics=dict(n_uploads='cardinality')))

    upload_name = metainfo.Quantity(
        type=str, categories=[MongoUploadMetadata, EditableUserMetadata],
        description='The user provided upload name',
        a_elasticsearch=Elasticsearch())

    upload_create_time = metainfo.Quantity(
        type=metainfo.Datetime, categories=[MongoUploadMetadata, EditableUserMetadata],
        description='The date and time when the upload was created in nomad',
        a_auth_level=AuthLevel.admin,
        a_elasticsearch=Elasticsearch(material_entry_type))

    entry_id = metainfo.Quantity(
        type=str,
        description='A persistent and globally unique identifier for the entry',
        categories=[MongoEntryMetadata, MongoSystemMetadata],
        a_elasticsearch=Elasticsearch(material_entry_type, metrics=dict(n_entries='cardinality')))

    entry_name = metainfo.Quantity(
        type=str,
        description='A brief human readable name for the entry.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type, _es_field='keyword'),
            Elasticsearch(
                material_entry_type, field='prefix',
                es_query='match_phrase_prefix', mapping='text', _es_field='')
        ])

    entry_type = metainfo.Quantity(
        type=str,
        description='The main schema definition. This is the name of the section used for data.',
        a_elasticsearch=Elasticsearch(material_entry_type))

    calc_id = metainfo.Quantity(
        type=str, description='Legacy field name, use `entry_id` instead.',
        derived=lambda entry: entry.entry_id,
        a_elasticsearch=Elasticsearch(material_entry_type))

    entry_hash = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=str,
        description='A raw file content based checksum/hash',
        categories=[MongoEntryMetadata])

    entry_create_time = metainfo.Quantity(
        type=metainfo.Datetime, categories=[MongoEntryMetadata, MongoSystemMetadata, EditableUserMetadata],
        description='The date and time when the entry was created in nomad',
        a_auth_level=AuthLevel.admin,
        a_elasticsearch=Elasticsearch(material_entry_type))

    last_edit_time = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=metainfo.Datetime, categories=[MongoEntryMetadata],
        description='The date and time the user metadata was last edited.')

    parser_name = metainfo.Quantity(
        type=str, categories=[MongoEntryMetadata, MongoSystemMetadata],
        description='The NOMAD parser used for the last processing',
        a_elasticsearch=Elasticsearch())

    mainfile = metainfo.Quantity(
        type=str, categories=[MongoEntryMetadata, MongoSystemMetadata],
        description='The path to the mainfile from the root directory of the uploaded files',
        a_elasticsearch=PathSearch())

    mainfile_key = metainfo.Quantity(
        type=str, categories=[MongoEntryMetadata, MongoSystemMetadata],
        description='''
            Key used to differentiate between different *child entries* of an entry.
            For parent entries and entries that do not have any children, the value should
            be empty.
        ''',
        a_elasticsearch=PathSearch())

    files = metainfo.Quantity(
        type=str, shape=['0..*'],
        description='''
            The paths to the files within the upload that belong to this entry.
            All files within the same directory as the entry's mainfile are considered the
            auxiliary files that belong to the entry.
        ''',
        a_elasticsearch=PathSearch())

    pid = metainfo.Quantity(
        type=str,
        description='''
            The unique, sequentially enumerated, integer PID that was used in the legacy
            NOMAD CoE. It allows to resolve URLs of the old NOMAD CoE Repository.
        ''',
        categories=[MongoEntryMetadata],
        a_elasticsearch=Elasticsearch(es_entry_type))

    raw_id = metainfo.Quantity(
        type=str,
        description='''
            The code specific identifier extracted from the entry's raw files by the parser,
            if supported.
        ''',
        a_elasticsearch=Elasticsearch(es_entry_type))

    external_id = metainfo.Quantity(
        type=str, categories=[MongoEntryMetadata, EditableUserMetadata],
        description='''
            A user provided external id. Usually the id for an entry in an external database
            where the data was imported from.
        ''',
        a_elasticsearch=Elasticsearch())

    published = metainfo.Quantity(
        type=bool, default=False,
        description='Indicates if the entry is published',
        categories=[MongoUploadMetadata],
        a_elasticsearch=Elasticsearch(material_entry_type))

    publish_time = metainfo.Quantity(
        type=metainfo.Datetime, categories=[MongoUploadMetadata, EditableUserMetadata],
        description='The date and time when the upload was published in nomad',
        a_auth_level=AuthLevel.admin,
        a_elasticsearch=Elasticsearch(material_entry_type))

    with_embargo = metainfo.Quantity(
        type=bool, default=False,
        categories=[MongoUploadMetadata, MongoSystemMetadata],
        description='Indicated if this entry is under an embargo',
        a_elasticsearch=Elasticsearch(material_entry_type))

    embargo_length = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=int, categories=[MongoUploadMetadata, EditableUserMetadata],
        description='The length of the requested embargo period, in months')

    license = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=str,
        description='''
            A short license description (e.g. CC BY 4.0), that refers to the
            license of this entry.
        ''',
        default='CC BY 4.0',
        categories=[MongoUploadMetadata, EditableUserMetadata],
        a_auth_level=AuthLevel.admin)

    processed = metainfo.Quantity(
        type=bool, default=False, categories=[MongoEntryMetadata, MongoSystemMetadata],
        description='Indicates that the entry is successfully processed.',
        a_elasticsearch=Elasticsearch())

    last_processing_time = metainfo.Quantity(
        type=metainfo.Datetime,
        description='The date and time of the last processing.',
        categories=[MongoEntryMetadata],
        a_elasticsearch=Elasticsearch())

    processing_errors = metainfo.Quantity(
        type=str, shape=['*'], description='Errors that occured during processing',
        a_elasticsearch=Elasticsearch())

    nomad_version = metainfo.Quantity(
        type=str,
        description='The NOMAD version used for the last processing',
        categories=[MongoEntryMetadata],
        a_elasticsearch=Elasticsearch())

    nomad_commit = metainfo.Quantity(
        type=str,
        description='The NOMAD commit used for the last processing',
        categories=[MongoEntryMetadata],
        a_elasticsearch=Elasticsearch())

    comment = metainfo.Quantity(
        type=str, categories=[MongoEntryMetadata, EditableUserMetadata],
        description='A user provided comment for this entry',
        a_elasticsearch=Elasticsearch(mapping='text'))

    references = metainfo.Quantity(
        type=str, shape=['0..*'], categories=[MongoEntryMetadata, EditableUserMetadata],
        description='User provided references (URLs) for this entry',
        a_elasticsearch=Elasticsearch())

    external_db = metainfo.Quantity(
        type=metainfo.MEnum('EELS Data Base', 'Materials Project', 'AFLOW', 'OQMD'),
        categories=[MongoUploadMetadata, EditableUserMetadata],
        description='The repository or external database where the original data resides',
        a_elasticsearch=Elasticsearch(material_entry_type))

    origin = metainfo.Quantity(
        type=str,
        description='''
            A short human readable description of the entries origin. Usually it is the
            handle of an external database/repository or the name of the main author.
        ''',
        derived=derive_origin,
        a_elasticsearch=Elasticsearch(material_entry_type))

    main_author = metainfo.Quantity(
        type=user_reference, categories=[MongoUploadMetadata, EditableUserMetadata],
        description='The main author of the entry',
        a_auth_level=AuthLevel.admin,
        a_elasticsearch=Elasticsearch(material_entry_type))

    coauthors = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=author_reference, shape=['0..*'], default=[], categories=[MongoUploadMetadata, EditableUserMetadata],
        description='''
            A user provided list of co-authors for the whole upload. These can view and edit the
            upload when in staging, and view it also if it is embargoed.
        ''')

    entry_coauthors = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=author_reference, shape=['0..*'], default=[], categories=[MongoEntryMetadata],
        description='''
            A user provided list of co-authors specific for this entry. This is a legacy field,
            for new uploads, coauthors should be specified on the upload level only.
        ''')

    reviewers = metainfo.Quantity(
        # Note: This attribute is not stored in ES
        type=user_reference, shape=['0..*'], default=[], categories=[MongoUploadMetadata, EditableUserMetadata],
        description='''
            A user provided list of reviewers. Reviewers can see the whole upload, also if
            it is unpublished or embargoed
        ''')

    authors = metainfo.Quantity(
        type=author_reference, shape=['0..*'],
        description='All authors (main author and co-authors)',
        derived=derive_authors,
        a_elasticsearch=Elasticsearch(material_entry_type, metrics=dict(n_authors='cardinality')))

    writers = metainfo.Quantity(
        type=user_reference, shape=['0..*'],
        description='All writers (main author, upload coauthors)',
        derived=lambda entry: ([entry.main_author] if entry.main_author is not None else []) + entry.coauthors,
        a_elasticsearch=Elasticsearch(material_entry_type))

    viewers = metainfo.Quantity(
        type=user_reference, shape=['0..*'],
        description='All viewers (main author, upload coauthors, and reviewers)',
        derived=lambda entry: ([entry.main_author] if entry.main_author is not None else []) + entry.coauthors + entry.reviewers,
        a_elasticsearch=Elasticsearch(material_entry_type))

    datasets = metainfo.Quantity(
        type=dataset_reference, shape=['0..*'], default=[],
        categories=[MongoEntryMetadata, EditableUserMetadata],
        description='A list of user curated datasets this entry belongs to.',
        a_elasticsearch=Elasticsearch(material_entry_type))

    optimade = metainfo.SubSection(
        sub_section=OptimadeEntry,
        description='Metadata used for the optimade API.',
        a_elasticsearch=Elasticsearch(es_entry_type))

    domain = metainfo.Quantity(
        type=metainfo.MEnum('dft', 'ems'),
        description='The material science domain',
        a_elasticsearch=Elasticsearch(material_entry_type))

    n_quantities = metainfo.Quantity(
        type=int, default=0, description='Number of metainfo quantities parsed from the entry.',
        a_elasticsearch=Elasticsearch(metrics=dict(n_quantities='sum')))

    quantities = metainfo.Quantity(
        type=str, shape=['0..*'],
        description='All quantities that are used by this entry.',
        a_elasticsearch=QuantitySearch())

    sections = metainfo.Quantity(
        type=str, shape=['*'],
        description='All sections that are present in this entry.',
        a_elasticsearch=Elasticsearch(material_entry_type))

    def apply_archvie_metadata(self, archive):
        quantities = set()
        sections = set()
        n_quantities = 0

        section_paths = {}

        def get_section_path(section):
            section_path = section_paths.get(section)
            if section_path is None:
                parent = section.m_parent
                if parent:
                    parent_path = get_section_path(parent)
                    if parent_path == '':
                        section_path = section.m_parent_sub_section.name
                    else:
                        section_path = f'{parent_path}.{section.m_parent_sub_section.name}'
                else:
                    section_path = ''
                section_paths[section] = section_path
                quantities.add(section_path)

            return section_path

        for section, property_def, _ in archive.m_traverse():
            sections.add(section.m_def)

            quantity_path = f'{get_section_path(section)}.{property_def.name}'
            quantities.add(quantity_path)
            n_quantities += 1

        self.quantities = list(quantities)
        self.quantities.sort()
        self.sections = [section.qualified_name() for section in sections]
        self.sections.sort()
        self.n_quantities = n_quantities


class EntryArchive(metainfo.MSection):
    m_def = metainfo.Section(label='Entry')

    entry_id = metainfo.Quantity(
        type=str, description='The unique primary id for this entry.',
        derived=lambda entry: entry.metadata.entry_id)

    run = metainfo.SubSection(sub_section=Run, repeats=True)
    measurement = metainfo.SubSection(sub_section=Measurement, repeats=True)

    data = metainfo.SubSection(sub_section=EntryData)

    workflow = metainfo.SubSection(sub_section=Workflow, repeats=True, categories=[FastAccess])
    metadata = metainfo.SubSection(
        sub_section=EntryMetadata, categories=[FastAccess],
        a_elasticsearch=Elasticsearch())

    processing_logs = metainfo.Quantity(
        type=Any, shape=['0..*'],
        description='The processing logs for this entry as a list of structlog entries.')

    results = metainfo.SubSection(
        sub_section=Results,
        categories=[FastAccess],
        a_elasticsearch=Elasticsearch(auto_include_subsections=True))

    tabular_tree = metainfo.SubSection(sub_section=TabularTree, repeats=False)
    definitions = metainfo.SubSection(sub_section=metainfo.Package)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click

from .cli import cli


@cli.command(help='Run parsing and normalizing locally.', name='parse')
@click.argument('MAINFILE', nargs=1, required=True, type=str)
@click.option('--show-archive', is_flag=True, default=False, help='Print the archive data.')
@click.option('--show-metadata', is_flag=True, default=False, help='Print the extracted repo metadata.')
@click.option('--skip-normalizers', is_flag=True, default=False, help='Do not run the normalizer.')
@click.option('--not-strict', is_flag=True, help='Do also match artificial parsers.')
@click.option('--parser', help='Skip matching and use the provided parser')
def _parse(mainfile, show_archive, show_metadata, skip_normalizers, not_strict, parser):
    import sys
    import json

    from nomad.client import parse, normalize_all

    kwargs = dict(strict=not not_strict, parser_name=parser)

    entry_archives = parse(mainfile, **kwargs)

    for entry_archive in entry_archives:
        if not skip_normalizers:
            normalize_all(entry_archive)
            entry_archive.metadata.apply_archvie_metadata(entry_archive)

        if show_archive:
            json.dump(entry_archive.m_to_dict(), sys.stdout, indent=2)

        if show_metadata:
            metadata = entry_archive.metadata
            metadata.apply_archvie_metadata(entry_archive)
            json.dump(metadata.m_to_dict(), sys.stdout, indent=4)
